'''
    ModernGL textures
'''

from .common import InvalidObject, TextureFilter
from .buffers import Buffer


class Texture:
    '''
        A Texture is an OpenGL object that contains one or more images that all
        have the same image format. A texture can be used in two ways. It can
        be the source of a texture access from a Shader, or it can be used
        as a render target.

        A Texture object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.texture` or :py:meth:`Context.depth_texture`
        to create one.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Texture.__new__(Texture)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Texture: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def repeat_x(self) -> bool:
        '''
            bool: The repeat_x of the texture.
        '''

        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self) -> bool:
        '''
            bool: The repeat_y of the texture.
        '''

        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def filter(self) -> TextureFilter:
        '''
            TextureFilter: The filter of the texture.
        '''

        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value.mglo

    @property
    def swizzle(self) -> str:
        '''
            str: The swizzle of the texture.
        '''

        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def width(self) -> int:
        '''
            int: The width of the texture.
        '''

        return self.mglo.width

    @property
    def height(self) -> int:
        '''
            int: The height of the texture.
        '''

        return self.mglo.height

    @property
    def size(self) -> tuple:
        '''
            tuple: The size of the texture.
        '''

        return (self.mglo.width, self.mglo.height)

    @property
    def components(self) -> int:
        '''
            int: The number of components of the texture.
        '''

        return self.mglo.components

    @property
    def samples(self) -> int:
        '''
            int: The number of samples of the texture.
        '''

        return self.mglo.samples

    @property
    def floats(self) -> bool:
        '''
            bool: Is the texture using floats?
        '''

        return self.mglo.floats

    @property
    def depth(self) -> bool:
        '''
            bool: Is the texture a depth texture?
        '''

        return self.mglo.depth

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def read(self, *, alignment=1) -> bytes:
        '''
            Read the content of the texture into a buffer.

            Keyword Args:
                alignment (int): The byte alignment of the pixels.

            Returns:
                bytes: the pixels
        '''

        return self.mglo.read(alignment)

    def read_into(self, buffer, *, alignment=1, write_offset=0) -> None:
        '''
            Read the content of the texture into a buffer.

            Args:
                buffer (bytearray): The buffer that will receive the pixels.

            Keyword Args:
                alignment (int): The byte alignment of the pixels.
                write_offset (int): The write offset.
        '''

        if isinstance(buffer, Buffer):
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(self, data, viewport=None, *, alignment=1) -> None:
        '''
            Update the content of the texture.
        '''

        if isinstance(data, Buffer):
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base=0, max_level=1000) -> None:
        '''
            Generate mipmaps.
        '''

        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0) -> None:
        '''
            Bind the texture.

            Args:
                location (int): The texture location.
                    Same as the integer value that is used for sampler2D
                    uniforms in the shaders. The value ``0`` will bind the
                    texture to the ``GL_TEXTURE0`` binding point.
        '''

        self.mglo.use(location)

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject


class Texture3D:
    '''
        A Texture is an OpenGL object that contains one or more images that all
        have the same image format. A texture can be used in two ways. It can
        be the source of a texture access from a Shader, or it can be used
        as a render target.

        A Texture3D object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.texture3d` to create one.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Texture.__new__(Texture)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Texture3D: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def repeat_x(self) -> bool:
        '''
            bool: The repeat_x of the texture.
        '''

        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self) -> bool:
        '''
            bool: The repeat_y of the texture.
        '''

        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def repeat_z(self) -> bool:
        '''
            bool: The repeat_z of the texture.
        '''

        return self.mglo.repeat_z

    @repeat_z.setter
    def repeat_z(self, value):
        self.mglo.repeat_z = value

    @property
    def filter(self) -> TextureFilter:
        '''
            TextureFilter: The filter of the texture.
        '''

        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value.mglo

    @property
    def swizzle(self) -> str:
        '''
            str: The swizzle of the texture.
        '''

        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def width(self) -> int:
        '''
            int: The width of the texture.
        '''

        return self.mglo.width

    @property
    def height(self) -> int:
        '''
            int: The height of the texture.
        '''

        return self.mglo.height

    @property
    def depth(self) -> int:
        '''
            int: The depth of the texture.
        '''

        return self.mglo.depth

    @property
    def size(self) -> tuple:
        '''
            tuple: The size of the texture.
        '''

        return (self.mglo.width, self.mglo.height, self.mglo.depth)

    @property
    def components(self) -> int:
        '''
            int: The number of components of the texture.
        '''

        return self.mglo.components

    @property
    def floats(self) -> bool:
        '''
            bool: Is the texture using floats?
        '''

        return self.mglo.floats

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def read(self, *, alignment=1) -> bytes:
        '''
            Read the content of the texture into a buffer.

            Args:
                buffer (bytearray): The buffer that will receive the pixels.
                viewport (tuple): The viewport.

            Keyword Args:
                alignment (int): The byte alignment of the pixels.

            Returns:
                bytes: the pixels
        '''

        return self.mglo.read(alignment)

    def read_into(self, buffer, *, alignment=1, write_offset=0) -> None:
        '''
            Read the content of the texture into a buffer.

            Args:
                buffer (bytearray): The buffer that will receive the pixels.
                viewport (tuple): The viewport.

            Keyword Args:
                alignment (int): The byte alignment of the pixels.
                write_offset (int): The write offset.
        '''

        if isinstance(buffer, Buffer):
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(self, data, viewport=None, *, alignment=1) -> None:
        '''
            Update the content of the texture.
        '''

        if isinstance(data, Buffer):
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base=0, max_level=1000) -> None:
        '''
            Generate mipmaps.
        '''

        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0) -> None:
        '''
            Bind the texture.

            Args:
                location (int): The texture location.
                    Same as the integer value that is used for sampler3D
                    uniforms in the shaders. The value ``0`` will bind the
                    texture to the ``GL_TEXTURE0`` binding point.
        '''

        self.mglo.use(location)

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject



import struct
import unittest

import ModernGL


class TestBuffer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        cls.vert = cls.ctx.vertex_shader('''
            #version 330

            in mat2 in_m;
            in vec2 in_v;

            out vec2 out_v;

            uniform float mult;

            void main() {
                out_v = in_m * in_v * mult;
            }
        ''')

        cls.prog = cls.ctx.program(cls.vert, ['out_v'])

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_1(self):
        buf_m = self.ctx.buffer(struct.pack('4f', 1, 1, 1, 2))
        buf_v = self.ctx.buffer(struct.pack('2f', 4, 7))
        res = self.ctx.buffer(reserve=buf_v.size)

        vao = self.ctx.vertex_array(self.prog, [
            (buf_m, '4f', ['in_m']),
            (buf_v, '2f', ['in_v']),
        ])

        self.prog.uniforms['mult'].value = 0.0
        vao.transform(res, ModernGL.POINTS)
        x, y = struct.unpack('2f', res.read())
        self.assertAlmostEqual(x, 0.0)
        self.assertAlmostEqual(y, 0.0)

        self.prog.uniforms['mult'].value = 1.0
        vao.transform(res, ModernGL.POINTS)
        x, y = struct.unpack('2f', res.read())
        self.assertAlmostEqual(x, 11.0)
        self.assertAlmostEqual(y, 18.0)

        self.prog.uniforms['mult'].value = 2.0
        vao.transform(res, ModernGL.POINTS)
        x, y = struct.unpack('2f', res.read())
        self.assertAlmostEqual(x, 22.0)
        self.assertAlmostEqual(y, 36.0)


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_viewport(self):
        x, y, width, height = self.ctx.default_framebuffer.viewport
        self.assertEqual(x, 0)
        self.assertEqual(y, 0)
        self.assertEqual(width, 0)
        self.assertEqual(height, 0)

    def test_size(self):
        width, height = self.ctx.default_framebuffer.size
        self.assertEqual(width, 0)
        self.assertEqual(height, 0)

    def test_samples(self):
        samples = self.ctx.default_framebuffer.samples
        self.assertEqual(samples, 0)

    def test_having_color_attachments(self):
        with self.assertRaisesRegex(ModernGL.Error, 'attachment'):
            self.ctx.default_framebuffer.color_attachments

    def test_having_depth_attachment(self):
        with self.assertRaisesRegex(ModernGL.Error, 'attachment'):
            self.ctx.default_framebuffer.depth_attachment


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        prog = cls.ctx.program([
            cls.ctx.vertex_shader('''
                #version 330

                in float v_in_1;
                in float v_in_2;
                in float v_in_3;
                out float v_out;

                void main() {
                    v_out = v_in_1 + v_in_2 + v_in_3;
                }
            '''),
        ], varyings=['v_out'])

        vbo1 = cls.ctx.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))
        vbo2 = cls.ctx.buffer(struct.pack('4f', 10.0, 20.0, 30.0, 40.0))
        vbo3 = cls.ctx.buffer(struct.pack('4f', 100.0, 200.0, 300.0, 400.0))

        vao1_content = [
            (vbo1, 'f', ['v_in_1']),
            (vbo2, 'f/i', ['v_in_2']),
        ]

        cls.vao1 = cls.ctx.vertex_array(prog, vao1_content)

        vao2_content = [
            (vbo1, 'f', ['v_in_1']),
            (vbo2, 'f/i', ['v_in_2']),
            (vbo3, 'f/r', ['v_in_3']),
        ]

        cls.vao2 = cls.ctx.vertex_array(prog, vao2_content)

        cls.res = cls.ctx.buffer(reserve=1024)

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_vertex_attrib_per_instance_1(self):
        self.vao1.transform(self.res, ModernGL.POINTS, vertices=4, instances=1)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 11.0)
        self.assertAlmostEqual(b, 12.0)
        self.assertAlmostEqual(c, 13.0)
        self.assertAlmostEqual(d, 14.0)

    def test_vertex_attrib_per_instance_2(self):
        self.vao1.transform(self.res, ModernGL.POINTS, vertices=2, instances=2)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 11.0)
        self.assertAlmostEqual(b, 12.0)
        self.assertAlmostEqual(c, 21.0)
        self.assertAlmostEqual(d, 22.0)

    def test_vertex_attrib_per_instance_3(self):
        self.vao1.transform(self.res, ModernGL.POINTS, vertices=1, instances=4)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 11.0)
        self.assertAlmostEqual(b, 21.0)
        self.assertAlmostEqual(c, 31.0)
        self.assertAlmostEqual(d, 41.0)

    def test_vertex_attrib_per_render_1(self):
        self.vao2.transform(self.res, ModernGL.POINTS, vertices=4, instances=1)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 111.0)
        self.assertAlmostEqual(b, 112.0)
        self.assertAlmostEqual(c, 113.0)
        self.assertAlmostEqual(d, 114.0)

    def test_vertex_attrib_per_render_2(self):
        self.vao2.transform(self.res, ModernGL.POINTS, vertices=2, instances=2)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 111.0)
        self.assertAlmostEqual(b, 112.0)
        self.assertAlmostEqual(c, 121.0)
        self.assertAlmostEqual(d, 122.0)

    def test_vertex_attrib_per_render_3(self):
        self.vao2.transform(self.res, ModernGL.POINTS, vertices=1, instances=4)

        a, b, c, d = struct.unpack('4f', self.res.read(16))

        self.assertAlmostEqual(a, 111.0)
        self.assertAlmostEqual(b, 121.0)
        self.assertAlmostEqual(c, 131.0)
        self.assertAlmostEqual(d, 141.0)


if __name__ == '__main__':
    unittest.main()



import importlib
import unittest


class TestCase(unittest.TestCase):

    def test_import_and_context(self):
        mgl = importlib.import_module('ModernGL')
        ctx = mgl.create_standalone_context(require=mgl.CORE_330)
        ctx.release()


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        cls.vert = cls.ctx.vertex_shader('''
            #version 330

            in vec2 in_v;
            out vec2 out_v;

            uniform Block1 {
                float x;
            };

            uniform Block2 {
                float y;
            };

            uniform Block3 {
                float z;
            };

            void main() {
                out_v = in_v * z + vec2(x, y);
            }
        ''')

        cls.prog = cls.ctx.program(cls.vert, ['out_v'])

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_1(self):
        buf_v = self.ctx.buffer(struct.pack('2f', 100.0, 1000.0))
        buf_u1 = self.ctx.buffer(struct.pack('f', 9.5))
        buf_u2 = self.ctx.buffer(struct.pack('f', 4.0))
        buf_u3 = self.ctx.buffer(struct.pack('f', 3.0))
        buf_u4 = self.ctx.buffer(struct.pack('f', 0.0))
        buf_r = self.ctx.buffer(reserve=buf_v.size)

        vao = self.ctx.vertex_array(self.prog, [
            (buf_v, '2f', ['in_v']),
        ])

        self.assertFalse('Block1' in self.prog.uniforms)
        self.assertFalse('Block2' in self.prog.uniforms)
        self.assertFalse('Block3' in self.prog.uniforms)

        self.assertFalse('x' in self.prog.uniforms)
        self.assertFalse('y' in self.prog.uniforms)
        self.assertFalse('z' in self.prog.uniforms)

        self.assertTrue('Block1' in self.prog.uniform_blocks)
        self.assertTrue('Block2' in self.prog.uniform_blocks)
        self.assertTrue('Block3' in self.prog.uniform_blocks)

        self.prog.uniform_blocks['Block1'].binding = 2
        self.prog.uniform_blocks['Block2'].binding = 4
        self.prog.uniform_blocks['Block3'].binding = 1

        buf_u1.bind_to_uniform_block(2)
        buf_u2.bind_to_uniform_block(4)
        buf_u4.bind_to_uniform_block(1)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, 9.5)
        self.assertAlmostEqual(b, 4.0)

        buf_u3.bind_to_uniform_block(1)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, 309.5)
        self.assertAlmostEqual(b, 3004.0)


if __name__ == '__main__':
    unittest.main()



'''
    The tkinter ModernGL connector
'''

# pylint: disable=too-many-ancestors

import struct

import ctypes
import tkinter


class ModernGLCanvasError(Exception):
    '''
        Custom error type for ModernGLCanvas
    '''


class ModernGLCanvas(tkinter.Frame):
    '''
        The tkinter ModernGL connector
    '''

    def __init__(self, root, width, height):
        tkinter.Frame.__init__(self, root, width=width, height=height, bg="", colormap="new")

        self._width = width
        self._height = height

        # typedef struct tagBITMAPINFOHEADER {
        #     DWORD biSize;
        #     LONG biWidth;
        #     LONG biHeight;
        #     WORD biPlanes;
        #     WORD biBitCount;
        #     DWORD biCompression;
        #     DWORD biSizeImage;
        #     LONG biXPelsPerMeter;
        #     LONG biYPelsPerMeter;
        #     DWORD biClrUsed;
        #     DWORD biClrImportant;
        # } BITMAPINFOHEADER;

        hbmpi = struct.pack('IIIHHIIIIII', 40, width, height, 1, 24, 0, 0, 0, 0, 0, 0)

        self._hwnd = self.winfo_id()

        if not self._hwnd:
            raise ModernGLCanvasError('cannot get the hwnd')

        self._hdc = ctypes.windll.user32.GetDC(self._hwnd)

        if not self._hdc:
            raise ModernGLCanvasError('cannot get the hdc')

        self._hmemdc = ctypes.windll.gdi32.CreateCompatibleDC(self._hdc)

        if not self._hmemdc:
            raise ModernGLCanvasError('cannot create device content')

        bits = ctypes.c_void_p()
        self._hbmp = ctypes.windll.gdi32.CreateDIBSection(self._hmemdc, hbmpi, 0, ctypes.byref(bits), 0, 0)

        if not self._hbmp:
            raise ModernGLCanvasError('cannot create bitmap section')

        if not bits.value:
            raise ModernGLCanvasError('cannot access the pixel data')

        array_type = (ctypes.c_char * width * height * 3)
        self._pixels = array_type.from_address(bits.value)

        sel = ctypes.windll.gdi32.SelectObject(self._hmemdc, self._hbmp)

        if not sel:
            raise ModernGLCanvasError('cannot select bitmap')

    @property
    def pixels(self):
        '''
            The pixels.
        '''

        return self._pixels

    def update(self) -> None:
        ctypes.windll.gdi32.BitBlt(self._hdc, 0, 0, self._width, self._height, self._hmemdc, 0, 0, 0x00CC0020)



'''
    ModernGL renderbuffers
'''

from .common import InvalidObject


class Renderbuffer:
    '''
        Renderbuffer objects are OpenGL objects that contain images.
        They are created and used specifically with :py:class:`Framebuffer` objects.
        They are optimized for use as render targets, while :py:class:`Texture` objects
        may not be, and are the logical choice when you do not need to sample
        from the produced image. If you need to resample, use Textures instead.
        Renderbuffer objects also natively accommodate multisampling.

        A Renderbuffer object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.renderbuffer` or :py:meth:`Context.depth_renderbuffer`
        to create one.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Renderbuffer.__new__(Renderbuffer)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Renderbuffer: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def width(self) -> int:
        '''
            int: The width of the renderbuffer.
        '''

        return self.mglo.width

    @property
    def height(self) -> int:
        '''
            int: The height of the renderbuffer.
        '''

        return self.mglo.height

    @property
    def size(self) -> tuple:
        '''
            tuple: The size of the renderbuffer.
        '''

        return (self.mglo.width, self.mglo.height)

    @property
    def samples(self) -> int:
        '''
            int: The samples of the renderbuffer.
        '''

        return self.mglo.samples

    @property
    def components(self) -> int:
        '''
            int: The components of the renderbuffer.
        '''

        return self.mglo.components

    @property
    def depth(self) -> bool:
        '''
            bool: Is the renderbuffer a depth renderbuffer?
        '''

        return self.mglo.depth

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject



'''
    ModernGL: PyOpenGL alternative
'''

import os

try:
    from .mgl import Error

except ImportError:

    if os.environ.get('READTHEDOCS') == 'True':
        from .mock import Error

    else:
        _IMPORT_ERROR = '\n'.join([
            'No implementation found for ModernGL',
            'Are you sure the source code is compiled properly?',
            'Hint: python3 setup.py build_ext -b .',
        ])

        raise ImportError(_IMPORT_ERROR)


from .common import (
    InvalidObject, Version, EnableFlag, Primitive, TextureFilter,
    POINTS, LINES, LINE_STRIP, LINE_LOOP, TRIANGLES, TRIANGLE_STRIP, TRIANGLE_FAN,
    LINES_ADJACENCY, LINE_STRIP_ADJACENCY, TRIANGLES_ADJACENCY, TRIANGLE_STRIP_ADJACENCY,
    CORE_330, CORE_400, CORE_410, CORE_420, CORE_430, CORE_440, CORE_450,
    BLEND, DEPTH_TEST, CULL_FACE, MULTISAMPLE,
    LINEAR, NEAREST, MIPMAP,
)

from .members import (
    Uniform, UniformMap, UniformBlock, UniformBlockMap,
    Varying, VaryingMap, Attribute, AttributeMap,
    Subroutine, SubroutineMap, SubroutineUniform, SubroutineUniformMap,
    ProgramStage,
)

from .vertex_arrays import (
    VertexArrayAttribute, VertexArray
)

from .context import Context, create_context, create_standalone_context
from .buffers import Buffer, BufferAccess, detect_format
from .programs import ComputeShader, Shader, Program
from .textures import Texture, Texture3D
from .renderbuffers import Renderbuffer
from .framebuffers import Framebuffer

__version__ = '4.1.11'

__all__ = [
    'InvalidObject', 'Version', 'EnableFlag', 'Primitive', 'Error', 'TextureFilter',
    'POINTS', 'LINES', 'LINE_STRIP', 'LINE_LOOP', 'TRIANGLES', 'TRIANGLE_STRIP', 'TRIANGLE_FAN',
    'LINES_ADJACENCY', 'LINE_STRIP_ADJACENCY', 'TRIANGLES_ADJACENCY', 'TRIANGLE_STRIP_ADJACENCY',
    'CORE_330', 'CORE_400', 'CORE_410', 'CORE_420', 'CORE_430', 'CORE_440', 'CORE_450',
    'BLEND', 'DEPTH_TEST', 'CULL_FACE', 'MULTISAMPLE',
    'LINEAR', 'NEAREST', 'MIPMAP',
    'Subroutine', 'SubroutineMap', 'SubroutineUniform', 'SubroutineUniformMap',
    'Uniform', 'UniformMap', 'UniformBlock', 'UniformBlockMap',
    'Varying', 'VaryingMap', 'Attribute', 'AttributeMap',
    'Program', 'ProgramStage', 'Shader', 'ComputeShader',
    'Texture', 'Texture3D', 'Framebuffer', 'Renderbuffer',
    'VertexArray', 'VertexArrayAttribute',
    'Buffer', 'BufferAccess',
    'Context',
    'create_context', 'create_standalone_context',
    'detect_format',
]



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    def test_1(self):
        self.assertEqual(repr(ModernGL.CORE_330), 'ModernGL.CORE_330')
        self.assertEqual(repr(ModernGL.TRIANGLES), 'ModernGL.TRIANGLES')
        self.assertEqual(repr(ModernGL.BLEND), 'ModernGL.BLEND')


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_vertex_shader(self):
        self.ctx.vertex_shader('''
            #version 330

            void main() {
                gl_Position = vec4(0, 0, 0, 0);
            }
        ''')

    def test_fragment_shader(self):
        self.ctx.fragment_shader('''
            #version 330

            out vec4 color;

            void main() {
                color = vec4(0, 0, 0, 0);
            }
        ''')


if __name__ == '__main__':
    unittest.main()



'''
    ModernGL buffers
'''

from .common import InvalidObject


class BufferAccess:
    '''
        BufferAccess objects are designed to access
        a :py:class:`Buffer` object's content inside a ``with`` statement.
        The buffer is mapped and unmapped only once.

        Use :py:meth:`Buffer.access` to get a BufferAccess object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = BufferAccess.__new__(BufferAccess)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('BufferAccess')

    def __repr__(self):
        return '<BufferAccess>'

    def __enter__(self):
        self.mglo.open()
        return self

    def __exit__(self, *args):
        self.mglo.close()

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    def open(self) -> None:
        '''
            Map the buffer.
            This method is called by ``__enter__``.
        '''

        self.mglo.open()

    def close(self) -> None:
        '''
            Unmap the buffer.
            This method is called by ``__exit__``.
        '''

        self.mglo.close()

    def read(self, size=-1, *, offset=0) -> bytes:
        '''
            Read the content.

            Args:
                size (int): The size. Value `-1` means all.

            Keyword Args:
                offset (int): The offset.

            Returns:
                bytes: binary data
        '''

        return self.mglo.read(size, offset)

    def read_into(self, buffer, size=-1, *, offset=0, write_offset=0) -> None:
        '''
            Read the content.

            Args:
                buffer (bytarray): The buffer that will receive the content.
                size (int): The size. Value `-1` means all.

            Keyword Args:
                offset (int): The read offset.
                write_offset (int): The write offset.

            Returns:
                bytes: binary data
        '''

        return self.mglo.read_into(buffer, size, offset, write_offset)

    def write(self, data, *, offset=0) -> None:
        '''
            Write the content.

            Args:
                size (int): The data.

            Keyword Args:
                offset (int): The offset.
        '''

        self.mglo.write(data, offset)

    @property
    def offset(self) -> int:
        '''
            int: The offset.
        '''

        return self.mglo.offset

    @property
    def size(self) -> int:
        '''
            int: The size.
        '''

        return self.mglo.size

    @property
    def readonly(self) -> bool:
        '''
            bool: Is readonly.
        '''

        return self.mglo.readonly


class Buffer:
    '''
        Buffer objects are OpenGL objects that store an array of unformatted memory
        allocated by the OpenGL context, (data allocated on the GPU).
        These can be used to store vertex data, pixel data retrieved from images
        or the framebuffer, and a variety of other things.

        A Buffer object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.buffer` to create one.

        Copy buffer content using :py:meth:`Context.copy_buffer`.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Buffer.__new__(Buffer)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Buffer: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def size(self) -> int:
        '''
            int: The size of the buffer.
        '''

        return self.mglo.size

    @property
    def dynamic(self) -> bool:
        '''
            bool: Is the buffer created with the dynamic flag?
        '''

        return self.mglo.dynamic

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def access(self, *, size=-1, offset=0, readonly=False) -> BufferAccess:
        '''
            Create a :py:class:`BufferAccess` object.

            Keyword Args:
                size (int): The size. Value `-1` means all.
                offset (int): The offset.
                readonly (bool): The readonly.

            Examples:

                Simple ``with`` statement::

                    # The buffer will be mapped once and accessed multiple times.

                    >>> with buffer.access() as access:
                    ...     access.read(...)
                    ...     access.write(...)
        '''

        return BufferAccess.new(self.mglo.access(size, offset, readonly))

    def read(self, size=-1, *, offset=0) -> bytes:
        '''
            Read the content.

            Args:
                size (int): The size. Value ``-1`` means all.

            Keyword Args:
                offset (int): The offset.

            Returns:
                bytes: The content of the buffer.
        '''

        return self.mglo.read(size, offset)

    def read_into(self, buffer, size=-1, *, offset=0, write_offset=0) -> None:
        '''
            Read the content into a buffer.

            Args:
                buffer (bytarray): The buffer that will receive the content.
                size (int): The size. Value ``-1`` means all.

            Keyword Args:
                offset (int): The read offset.
                write_offset (int): The write offset.

            Returns:
                bytes: The content of the buffer.
        '''

        return self.mglo.read_into(buffer, size, offset, write_offset)

    def write(self, data, *, offset=0) -> None:
        '''
            Write the content.

            Args:
                data (bytes): The data.

            Keyword Args:
                offset (int): The offset.
        '''

        self.mglo.write(data, offset)

    def clear(self, size=-1, *, offset=0, chunk=None) -> bytes:
        '''
            Clear the content.

            Args:
                size (int): The size. Value ``-1`` means all.

            Keyword Args:
                offset (int): The offset.
                chunk (bytes): The chunk to use repeatedly.
        '''

        return self.mglo.clear(size, offset, chunk)

    def orphan(self) -> None:
        '''
            Orphan the buffer.

            It is also called buffer re-specification.

            Reallocate the buffer object before you start modifying it.

            Since allocating storage is likely faster than the implicit synchronization,
            you gain significant performance advantages over synchronization.

            The old storage will still be used by the OpenGL commands that have been sent previously.
            It is likely that the GL driver will not be doing any allocation at all,
            but will just be pulling an old free block off the unused buffer queue and use it,
            so it is likely to be very efficient.

            Examples:

                Simple orphaning example::

                    # For simplicity the VertexArray creation is omitted

                    >>> vbo = ctx.buffer(reserve=1024)

                    # Fill the buffer

                    >>> vbo.write(some_temorary_data)

                    # Issue a render call that uses the vbo

                    >>> vao.render(...)

                    # Orphan the buffer

                    >>> vbo.orphan()

                    # Issue another render call without waiting for the previous one

                    >>> vbo.write(some_temorary_data)
                    >>> vao.render(...)
        '''

        self.mglo.orphan()

    def bind_to_uniform_block(self, binding=0) -> None:
        '''
            Bind the buffer to a uniform block.

            Args:
                binding (int): The uniform block binding.
        '''

        self.mglo.bind_to_uniform_block(binding)

    def bind_to_storage_buffer(self, binding=0) -> None:
        '''
            Bind the buffer to a shader storage buffer.

            Args:
                binding (int): The shader storage binding.
        '''

        self.mglo.bind_to_storage_buffer(binding)

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject


def detect_format(program, attributes) -> str:
    '''
        Detect format for vertex attributes.
        The format returned does not contain padding.

        Args:
            program (Program): The program.
            attributes (list): A list of attribute names.

        Returns:
            str: The tightly packed format for the attributes.
    '''

    def fmt(attr):
        '''
            For internal use only.
        '''

        return attr.array_length * attr.dimension, attr.shape

    return ''.join('%d%s' % fmt(program.attributes[a]) for a in attributes)



'''
    ModernGL programs
'''

# pylint: disable=too-few-public-methods

from .common import InvalidObject, Primitive
from .members import UniformMap, UniformBlockMap, VaryingMap, AttributeMap, ProgramStage


class ComputeShader:
    '''
        A Compute Shader is a Shader Stage that is used entirely for computing arbitrary information.
        While it can do rendering, it is generally used for tasks not directly related to drawing.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = ComputeShader.__new__(ComputeShader)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<ComputeShader: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def source(self) -> str:
        '''
            str: The source code of the compute shader.
        '''

        return self.mglo.source

    @property
    def uniforms(self) -> UniformMap:
        '''
            :py:class:`UniformMap`: The uniforms of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`Uniform` objects by name.

            Examples:

                Set the value of the uniforms::

                    # uniform vec3 eye_pos;
                    >>> program.uniforms['eye_pos'].value = (10.0, 20.0, 0.0)

                    # uniform sampler2D my_textures[3];
                    >>> program.uniforms['my_texture'].value = [0, 3, 2]

                    # The values of `my_textures` will be:
                    my_textures[0] = GL_TEXTURE0             # GL_TEXTURE0
                    my_textures[1] = GL_TEXTURE0 + 3         # GL_TEXTURE3
                    my_textures[2] = GL_TEXTURE0 + 2         # GL_TEXTURE2

                Get information about the uniforms::

                    >>> program.uniforms['eye_pos'].location
                    0

                    >>> program.uniforms['eye_pos'].dimension
                    3

                    >>> program.uniforms['eye_pos'].value
                    (10.0, 20.0, 0.0)

                    >>> program.uniforms['my_textures'].dimension
                    1

                    >>> program.uniforms['my_textures'].array_length
                    3
        '''

        return UniformMap.new(self.mglo.uniforms)

    @property
    def uniform_blocks(self) -> UniformBlockMap:
        '''
            :py:class:`UniformBlockMap`: The uniform blocks of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`UniformBlock` objects by name.

            Examples:

                Get the location of the uniform block::

                    # uniform custom_material {
                    #     ...
                    # };

                    >>> program.uniform_blocks['custom_material'].location
                    16
        '''

        return UniformBlockMap.new(self.mglo.uniform_blocks)

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def run(self, group_x=1, group_y=1, group_z=1) -> None:
        '''
            Run the compute shader.

            Args:
                group_x (int): The number of work groups to be launched in the X dimension.
                group_y (int): The number of work groups to be launched in the Y dimension.
                group_z (int): The number of work groups to be launched in the Z dimension.
        '''

        return self.mglo.run(group_x, group_y, group_z)

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject


class Shader:
    '''
        Shader objects represent compiled GLSL code for a single shader stage.

        A Shader object cannot be instantiated directly, it requires a context.
        Use the following methods to create one:

            - :py:meth:`Context.vertex_shader`
            - :py:meth:`Context.fragment_shader`
            - :py:meth:`Context.geometry_shader`
            - :py:meth:`Context.tess_evaluation_shader`
            - :py:meth:`Context.tess_control_shader`
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Shader.__new__(Shader)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Shader: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def source(self) -> str:
        '''
            str: The source code of the shader.
        '''

        return self.mglo.source

    @property
    def typename(self) -> str:
        '''
            str: The type of the shader.
            The return value is a string.

            - ``'vertex_shader'``
            - ``'fragment_shader'``
            - ``'geometry_shader'``
            - ``'tess_evaluation_shader'``
            - ``'tess_control_shader'``
        '''

        return self.mglo.typename

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject


class Program:
    '''
        A Program object represents fully processed executable code
        in the OpenGL Shading Language, for one or more Shader stages.

        In ModernGL, a Program object can be assigned to :py:class:`VertexArray` objects.
        The VertexArray object  is capable of binding the Program object once the
        :py:meth:`VertexArray.render` or :py:meth:`VertexArray.transform` is called.

        Program objects has no method called ``use()``, VertexArrays encapsulate this mechanism.

        A Program object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.program` to create one.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Program.__new__(Program)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Program: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def uniforms(self) -> UniformMap:
        '''
            :py:class:`UniformMap`: The uniforms of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`Uniform` objects by name.

            Examples:

                Set the value of the uniforms::

                    # uniform vec3 eye_pos;
                    >>> program.uniforms['eye_pos'].value = (10.0, 20.0, 0.0)

                    # uniform sampler2D my_textures[3];
                    >>> program.uniforms['my_texture'].value = [0, 3, 2]

                    # The values of `my_textures` will be:
                    my_textures[0] = GL_TEXTURE0             # GL_TEXTURE0
                    my_textures[1] = GL_TEXTURE0 + 3         # GL_TEXTURE3
                    my_textures[2] = GL_TEXTURE0 + 2         # GL_TEXTURE2

                Get information about the uniforms::

                    >>> program.uniforms['eye_pos'].location
                    0

                    >>> program.uniforms['eye_pos'].dimension
                    3

                    >>> program.uniforms['eye_pos'].value
                    (10.0, 20.0, 0.0)

                    >>> program.uniforms['my_textures'].dimension
                    1

                    >>> program.uniforms['my_textures'].array_length
                    3
        '''

        return UniformMap.new(self.mglo.uniforms)

    @property
    def uniform_blocks(self) -> UniformBlockMap:
        '''
            :py:class:`UniformBlockMap`: The uniform blocks of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`UniformBlock` objects by name.

            Examples:

                Get the location of the uniform block::

                    # uniform custom_material {
                    #     ...
                    # };

                    >>> program.uniform_blocks['custom_material'].location
                    16
        '''

        return UniformBlockMap.new(self.mglo.uniform_blocks)

    @property
    def attributes(self) -> AttributeMap:
        '''
            :py:class:`AttributeMap`: The attributes of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`Attribute` objects by name.

            Examples:

                Set the default value for the attributes::

                    # in vec3 normal;
                    >>> program.attributes['normal'].default = (0.0, 0.0, 1.0)

                    # in vec2 texture_coordinates[3];
                    >>> program.attributes['texture_coordinates'].default = [
                    ...     (0.0, 0.0),
                    ...     (0.0, 0.0),
                    ...     (0.0, 0.0),
                    ... ]

                Get information about the attributes::

                    >>> program.attributes['normal'].default
                    (0.0, 0.0, 1.0)

                    >>> program.attributes['normal'].dimension
                    3

                    >>> program.attributes['texture_coordinates'].default
                    [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)]

                    >>> program.attributes['texture_coordinates'].array_length
                    3

                    >>> program.attributes['texture_coordinates'].dimension
                    2
        '''

        return AttributeMap.new(self.mglo.attributes)

    @property
    def varyings(self) -> VaryingMap:
        '''
            :py:class:`VaryingMap`: The varyings of the program.
            The return value is a dictinary like object.
            It can be used to access :py:class:`Varying` objects by name.

            The only reason varyings were added to allow verifying
            varyings programatically, they do not hold any other information.

            Examples:

                Check varyings::

                    >>> 'vertex_out' in program.varyings
                    True
        '''

        return VaryingMap.new(self.mglo.varyings)

    @property
    def geometry_input(self) -> Primitive:
        '''
            :py:class:`Primitive`: The geometry input primitive.
            The GeometryShader's input primitive if the GeometryShader exists.
            The geometry input primitive will be used for validation.
        '''

        return self.mglo.geometry_input

    @property
    def geometry_output(self) -> Primitive:
        '''
            :py:class:`Primitive`: The geometry output primitive.
            The GeometryShader's output primitive if the GeometryShader exists.
        '''

        return self.mglo.geometry_output

    @property
    def geometry_vertices(self) -> int:
        '''
            :obj:`int`: The maximum number of vertices that
            the geometry shader will output.
        '''

        return self.mglo.geometry_vertices

    @property
    def vertex_shader(self) -> ProgramStage:
        '''
            :py:class:`ProgramStage`: The vertex shader program stage.

            The return value is NOT a Shader.
        '''

        stage = self.mglo.vertex_shader

        if stage is not None:
            stage = ProgramStage.new(stage)

        return stage

    @property
    def fragment_shader(self) -> ProgramStage:
        '''
            :py:class:`ProgramStage`: The fragment shader program stage.

            The return value is NOT a Shader.
        '''

        stage = self.mglo.fragment_shader

        if stage is not None:
            stage = ProgramStage.new(stage)

        return stage

    @property
    def geometry_shader(self) -> ProgramStage:
        '''
            :py:class:`ProgramStage`: The geometry shader program stage.

            The return value is NOT a Shader.
        '''

        stage = self.mglo.geometry_shader

        if stage is not None:
            stage = ProgramStage.new(stage)

        return stage

    @property
    def tess_evaluation_shader(self) -> ProgramStage:
        '''
            :py:class:`ProgramStage`: The tesselation evaluation shader program stage.

            The return value is NOT a Shader.
        '''

        stage = self.mglo.tess_evaluation_shader

        if stage is not None:
            stage = ProgramStage.new(stage)

        return stage

    @property
    def tess_control_shader(self) -> ProgramStage:
        '''
            :py:class:`ProgramStage`: The tesselation control shader program stage.

            The return value is NOT a Shader.
        '''

        stage = self.mglo.tess_control_shader

        if stage is not None:
            stage = ProgramStage.new(stage)

        return stage

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_program(self):

        vertex_shader = self.ctx.vertex_shader('''
            #version 330

            uniform vec2 pos;
            uniform float scale;

            in vec2 vert;
            out vec2 v_vert;

            void main() {
                gl_Position = vec4(pos + vert * scale, 0.0, 1.0);
                v_vert = vert;
            }
        ''')

        fragment_shader = self.ctx.fragment_shader('''
            #version 330

            in vec2 v_vert;
            out vec4 color;

            void main() {
                color = vec4(v_vert, 0.0, 1.0);
            }
        ''')

        program = self.ctx.program([vertex_shader, fragment_shader])

        self.assertIsNot(program.vertex_shader, None)
        self.assertIsNot(program.fragment_shader, None)
        self.assertIs(program.geometry_shader, None)
        self.assertIs(program.tess_evaluation_shader, None)
        self.assertIs(program.tess_control_shader, None)

        self.assertIs(program.geometry_input, None)
        self.assertIs(program.geometry_output, None)

        self.assertEqual(len(program.uniforms), 2)
        self.assertEqual(len(program.attributes), 1)

        self.assertEqual(len(program.varyings), 0)
        self.assertEqual(len(program.uniform_blocks), 0)

        self.assertTrue('pos' in program.uniforms)
        self.assertTrue('scale' in program.uniforms)
        self.assertTrue('vert' not in program.uniforms)

        self.assertTrue('vert' in program.attributes)
        self.assertTrue('pos' not in program.attributes)


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_read_1(self):
        size = (4, 4)
        rbo1 = self.ctx.renderbuffer(size, floats=False)
        rbo2 = self.ctx.renderbuffer(size)
        rbo3 = self.ctx.renderbuffer(size)
        fbo1 = self.ctx.framebuffer(rbo1)
        fbo2 = self.ctx.framebuffer(rbo2)
        fbo3 = self.ctx.framebuffer(rbo3)
        fbo4 = self.ctx.framebuffer([rbo1, rbo2, rbo3])

        fbo4.clear(0.0, 0x99 / 255, 0xb2 / 255, 1.0)

        pixels1 = fbo1.read(components=4, floats=False)
        self.assertEqual(pixels1, b'\x00\x99\xb2\xff' * 16)

        pixels2 = fbo2.read(components=4, floats=False)
        self.assertEqual(pixels2, b'\x00\x99\xb2\xff' * 16)

        pixels3 = fbo3.read(components=4, floats=False)
        self.assertEqual(pixels3, b'\x00\x99\xb2\xff' * 16)

    def test_read_2(self):
        size = (4, 4)
        rbo1 = self.ctx.renderbuffer(size, floats=False)
        rbo2 = self.ctx.renderbuffer(size)
        rbo3 = self.ctx.renderbuffer(size)
        fbo1 = self.ctx.framebuffer(rbo1)
        fbo2 = self.ctx.framebuffer(rbo2)
        fbo3 = self.ctx.framebuffer(rbo3)
        fbo4 = self.ctx.framebuffer([rbo1, rbo2, rbo3])

        fbo1.clear(1.0, 0.0, 0.0, 1.0)
        fbo2.clear(0.0, 1.0, 0.0, 1.0)
        fbo3.clear(0.0, 0.0, 1.0, 1.0)

        pixels1 = fbo4.read(attachment=0, components=4, floats=False)
        self.assertEqual(pixels1, b'\xff\x00\x00\xff' * 16)

        pixels2 = fbo4.read(attachment=1, components=4, floats=False)
        self.assertEqual(pixels2, b'\x00\xff\x00\xff' * 16)

        pixels3 = fbo4.read(attachment=2, components=4, floats=False)
        self.assertEqual(pixels3, b'\x00\x00\xff\xff' * 16)

    def test_read_into(self):
        size = (4, 4)
        rbo1 = self.ctx.renderbuffer(size, floats=False)
        rbo2 = self.ctx.renderbuffer(size)
        rbo3 = self.ctx.renderbuffer(size)
        fbo1 = self.ctx.framebuffer(rbo1)
        fbo2 = self.ctx.framebuffer(rbo2)
        fbo3 = self.ctx.framebuffer(rbo3)
        fbo4 = self.ctx.framebuffer([rbo1, rbo2, rbo3])

        fbo1.clear(1.0, 0.0, 0.0, 1.0)
        fbo2.clear(0.0, 1.0, 0.0, 1.0)
        fbo3.clear(0.0, 0.0, 1.0, 1.0)

        pixels1 = bytearray(64)
        fbo4.read_into(pixels1, attachment=0, components=4, floats=False)
        self.assertEqual(bytes(pixels1), b'\xff\x00\x00\xff' * 16)

        pixels2 = bytearray(64)
        fbo4.read_into(pixels2, attachment=1, components=4, floats=False)
        self.assertEqual(bytes(pixels2), b'\x00\xff\x00\xff' * 16)

        pixels3 = bytearray(64)
        fbo4.read_into(pixels3, attachment=2, components=4, floats=False)
        self.assertEqual(bytes(pixels3), b'\x00\x00\xff\xff' * 16)


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_vao_attribs(self):

        vertex_shader = self.ctx.vertex_shader('''
            #version 330

            in vec4 in_vert;
            out vec4 out_vert;

            void main() {
                out_vert = in_vert;
            }
        ''')

        program = self.ctx.program(vertex_shader, ['out_vert'])

        vbo = self.ctx.buffer(struct.pack('12f', 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))
        res = self.ctx.buffer(reserve=struct.calcsize('8f'))

        vao = self.ctx.simple_vertex_array(program, vbo, ['in_vert'])
        vao.transform(res, ModernGL.POINTS, 2)

        for a, b in zip(struct.unpack('8f', res.read()), (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0)):
            self.assertAlmostEqual(a, b)

        vao.attributes['in_vert'].bind(vbo, 4, 0, 0)
        vao.transform(res, ModernGL.POINTS, 2)

        for a, b in zip(struct.unpack('8f', res.read()), (2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0)):
            self.assertAlmostEqual(a, b)

        vao.attributes['in_vert'].bind(vbo, 8, 20, 0)
        vao.transform(res, ModernGL.POINTS, 2)

        for a, b in zip(struct.unpack('8f', res.read()), (3.0, 4.0, 5.0, 6.0, 8.0, 9.0, 10.0, 11.0)):
            self.assertAlmostEqual(a, b)

        vao.attributes['in_vert'].bind(vbo, 12, 0, 1)
        vao.transform(res, ModernGL.POINTS, 2)

        for a, b in zip(struct.unpack('8f', res.read()), (4.0, 5.0, 6.0, 7.0, 4.0, 5.0, 6.0, 7.0)):
            self.assertAlmostEqual(a, b)


if __name__ == '__main__':
    unittest.main()



'''
    ModernGL program members
'''

# pylint: disable=too-few-public-methods


class Uniform:
    '''
        A uniform is a global GLSL variable declared with the "uniform" storage qualifier.
        These act as parameters that the user of a shader program can pass to that program.

        In ModernGL, Uniforms can be accessed using :py:attr:`Program.uniforms`
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Uniform.__new__(Uniform)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('Uniform')

    def __repr__(self):
        return '<Uniform: %d>' % self.location

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: The name of the uniform.
            The name does not contain leading `[0]`.
            The name may contain `[ ]` when the uniform is part of a struct.
        '''

        return self.mglo.name

    @property
    def location(self) -> int:
        '''
            int: The location of the uniform.
            The location holds the value returned by the glGetUniformLocation.
            To set the value of the uniform use the :py:attr:`value` instead.
        '''

        return self.mglo.location

    @property
    def dimension(self) -> int:
        '''
            int: The dimension of the uniform.

            +-----------------+-----------+
            | GLSL type       | dimension |
            +=================+===========+
            | sampler2D       | 1         |
            +-----------------+-----------+
            | sampler2DCube   | 1         |
            +-----------------+-----------+
            | sampler2DShadow | 1         |
            +-----------------+-----------+
            | bool            | 1         |
            +-----------------+-----------+
            | bvec2           | 2         |
            +-----------------+-----------+
            | bvec3           | 3         |
            +-----------------+-----------+
            | bvec4           | 4         |
            +-----------------+-----------+
            | int             | 1         |
            +-----------------+-----------+
            | ivec2           | 2         |
            +-----------------+-----------+
            | ivec3           | 3         |
            +-----------------+-----------+
            | ivec4           | 4         |
            +-----------------+-----------+
            | uint            | 1         |
            +-----------------+-----------+
            | uvec2           | 2         |
            +-----------------+-----------+
            | uvec3           | 3         |
            +-----------------+-----------+
            | uvec4           | 4         |
            +-----------------+-----------+
            | float           | 1         |
            +-----------------+-----------+
            | vec2            | 2         |
            +-----------------+-----------+
            | vec3            | 3         |
            +-----------------+-----------+
            | vec4            | 4         |
            +-----------------+-----------+
            | double          | 1         |
            +-----------------+-----------+
            | dvec2           | 2         |
            +-----------------+-----------+
            | dvec3           | 3         |
            +-----------------+-----------+
            | dvec4           | 4         |
            +-----------------+-----------+
            | mat2            | 4         |
            +-----------------+-----------+
            | mat2x3          | 6         |
            +-----------------+-----------+
            | mat2x4          | 8         |
            +-----------------+-----------+
            | mat3x2          | 6         |
            +-----------------+-----------+
            | mat3            | 9         |
            +-----------------+-----------+
            | mat3x4          | 12        |
            +-----------------+-----------+
            | mat4x2          | 8         |
            +-----------------+-----------+
            | mat4x3          | 12        |
            +-----------------+-----------+
            | mat4            | 16        |
            +-----------------+-----------+
            | dmat2           | 4         |
            +-----------------+-----------+
            | dmat2x3         | 6         |
            +-----------------+-----------+
            | dmat2x4         | 8         |
            +-----------------+-----------+
            | dmat3x2         | 6         |
            +-----------------+-----------+
            | dmat3           | 9         |
            +-----------------+-----------+
            | dmat3x4         | 12        |
            +-----------------+-----------+
            | dmat4x2         | 8         |
            +-----------------+-----------+
            | dmat4x3         | 12        |
            +-----------------+-----------+
            | dmat4           | 16        |
            +-----------------+-----------+
        '''

        return self.mglo.dimension

    @property
    def array_length(self) -> int:
        '''
            int: The length of the array of the uniform.
            The array_length is `1` for non array uniforms.
        '''

        return self.mglo.array_length

    @property
    def value(self):
        '''
            The value of the uniform.
            Reading the value of the uniform may force the GPU to sync.

            The value must be a tuple for non array uniforms.
            The value must be a list of tuples for array uniforms.
        '''

        return self.mglo.value

    @value.setter
    def value(self, value):
        self.mglo.value = value

    def read(self) -> bytes:
        '''
            Read the value of the uniform.
        '''

        return self.mglo.read()

    def write(self, value) -> None:
        '''
            Write the value of the uniform.
        '''

        return self.mglo.write(bytes(value))


class UniformMap:
    '''
        UniformMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = UniformMap.__new__(UniformMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('UniformMap')

    def __getitem__(self, key) -> Uniform:
        return Uniform.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class UniformBlock:
    '''
        UniformBlock
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = UniformBlock.__new__(UniformBlock)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('UniformBlock')

    def __repr__(self):
        return '<UniformBlock: %d>' % self.index

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: name
        '''

        return self.mglo.name

    @property
    def index(self) -> int:
        '''
            int: The index of the uniform block.
        '''

        return self.mglo.index

    @property
    def size(self) -> int:
        '''
            int: The size of the uniform block.
        '''

        return self.mglo.size

    @property
    def binding(self) -> int:
        '''
            int: The binding of the uniform block.
        '''

        return self.mglo.binding

    @binding.setter
    def binding(self, value):
        self.mglo.binding = value


class UniformBlockMap:
    '''
        UniformBlockMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = UniformBlockMap.__new__(UniformBlockMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('UniformBlockMap')

    def __getitem__(self, key) -> UniformBlock:
        return UniformBlock.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class Varying:
    '''
        Varying
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Varying.__new__(Varying)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('Varying')

    def __repr__(self):
        return '<Varying: %d>' % self.number

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: The name of the varying.
        '''

        return self.mglo.name

    @property
    def number(self) -> int:
        '''
            int: The number of the varying.
        '''

        return self.mglo.number


class VaryingMap:
    '''
        VaryingMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = VaryingMap.__new__(VaryingMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('VaryingMap')

    def __getitem__(self, key) -> Varying:
        return Varying.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class Attribute:
    '''
        Attribute
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Attribute.__new__(Attribute)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('Attribute')

    def __repr__(self):
        return '<Attribute: %d>' % self.location

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: The attribute name.
            The name will be filtered to have no array syntax on it's end.
            Attribute name without ``'[0]'`` ending if any.
        '''

        return self.mglo.name

    @property
    def location(self) -> int:
        '''
            int: The location of the attribute.
            The result of the glGetAttribLocation.
        '''

        return self.mglo.location

    @property
    def array_length(self) -> int:
        '''
            int: If the attribute is an array the array_length
            is the length of the array otherwise `1`.
        '''

        return self.mglo.array_length

    @property
    def dimension(self) -> int:
        '''
            int: The attribute dimension.

            +-----------+-----------+
            | GLSL type | dimension |
            +===========+===========+
            | int       | 1         |
            +-----------+-----------+
            | ivec2     | 2         |
            +-----------+-----------+
            | ivec3     | 3         |
            +-----------+-----------+
            | ivec4     | 4         |
            +-----------+-----------+
            | uint      | 1         |
            +-----------+-----------+
            | uvec2     | 2         |
            +-----------+-----------+
            | uvec3     | 3         |
            +-----------+-----------+
            | uvec4     | 4         |
            +-----------+-----------+
            | float     | 1         |
            +-----------+-----------+
            | vec2      | 2         |
            +-----------+-----------+
            | vec3      | 3         |
            +-----------+-----------+
            | vec4      | 4         |
            +-----------+-----------+
            | double    | 1         |
            +-----------+-----------+
            | dvec2     | 2         |
            +-----------+-----------+
            | dvec3     | 3         |
            +-----------+-----------+
            | dvec4     | 4         |
            +-----------+-----------+
            | mat2      | 4         |
            +-----------+-----------+
            | mat2x3    | 6         |
            +-----------+-----------+
            | mat2x4    | 8         |
            +-----------+-----------+
            | mat3x2    | 6         |
            +-----------+-----------+
            | mat3      | 9         |
            +-----------+-----------+
            | mat3x4    | 12        |
            +-----------+-----------+
            | mat4x2    | 8         |
            +-----------+-----------+
            | mat4x3    | 12        |
            +-----------+-----------+
            | mat4      | 16        |
            +-----------+-----------+
            | dmat2     | 4         |
            +-----------+-----------+
            | dmat2x3   | 6         |
            +-----------+-----------+
            | dmat2x4   | 8         |
            +-----------+-----------+
            | dmat3x2   | 6         |
            +-----------+-----------+
            | dmat3     | 9         |
            +-----------+-----------+
            | dmat3x4   | 12        |
            +-----------+-----------+
            | dmat4x2   | 8         |
            +-----------+-----------+
            | dmat4x3   | 12        |
            +-----------+-----------+
            | dmat4     | 16        |
            +-----------+-----------+
        '''

        return self.mglo.dimension

    @property
    def shape(self) -> str:
        '''
            str: The shape is a single character, representing the scalar type of the attriute.

            +---------+--------------------------+
            | shape   | GLSL types               |
            +=========+==========================+
            | ``'i'`` | int                      |
            |         +--------------------------+
            |         | ivec2 ivec3 ivec4        |
            +---------+--------------------------+
            | ``'I'`` | uint                     |
            |         +--------------------------+
            |         | uvec2 uvec3 uvec4        |
            +---------+--------------------------+
            | ``'f'`` | float                    |
            |         +--------------------------+
            |         | vec2 vec3 vec4           |
            |         +--------------------------+
            |         | mat2 mat3 mat4           |
            |         +--------------------------+
            |         | mat2x3 mat2x4 mat3x4     |
            |         | mat4x2 mat4x2 mat4x3     |
            +---------+--------------------------+
            | ``'d'`` | double                   |
            |         +--------------------------+
            |         | dvec2 dvec3 dvec4        |
            |         +--------------------------+
            |         | dmat2 dmat3 dmat4        |
            |         +--------------------------+
            |         | dmat2x3 dmat2x4 dmat3x4  |
            |         | dmat4x2 dmat4x2 dmat4x3  |
            +---------+--------------------------+
        '''

        return self.mglo.shape


class AttributeMap:
    '''
        AttributeMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = AttributeMap.__new__(AttributeMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('AttributeMap')

    def __getitem__(self, key) -> Attribute:
        return Attribute.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class Subroutine:
    '''
        Subroutine
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Subroutine.__new__(Subroutine)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('Subroutine')

    def __int__(self):
        return self.index

    def __repr__(self):
        return '<Subroutine: %d>' % self.index

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: The name of the subroutine.
        '''

        return self.mglo.name

    @property
    def index(self) -> int:
        '''
            int: The index of the subroutine.
        '''

        return self.mglo.index


class SubroutineMap:
    '''
        SubroutineMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = SubroutineMap.__new__(SubroutineMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('SubroutineMap')

    def __getitem__(self, key) -> Subroutine:
        return Subroutine.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class SubroutineUniform:
    '''
        SubroutineUniform
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = SubroutineUniform.__new__(SubroutineUniform)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('SubroutineUniform')

    def __repr__(self):
        return '<SubroutineUniform: %d>' % self.location

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def name(self) -> str:
        '''
            str: The name of the subroutine uniform.
        '''

        return self.mglo.name

    @property
    def location(self) -> int:
        '''
            int: The location of the subroutine uniform.
        '''

        return self.mglo.location


class SubroutineUniformMap:
    '''
        SubroutineUniformMap is a dictionary like object.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = SubroutineUniformMap.__new__(SubroutineUniformMap)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('SubroutineUniformMap')

    def __getitem__(self, key) -> SubroutineUniform:
        return SubroutineUniform.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


class ProgramStage:
    '''
        ProgramStage
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = ProgramStage.__new__(ProgramStage)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('ProgramStage')

    def __repr__(self):
        return '<ProgramStage>'

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def subroutines(self) -> SubroutineMap:
        '''
            SubroutineMap: The subroutines of the program stage.
        '''

        return SubroutineMap.new(self.mglo.subroutines)

    @property
    def subroutine_uniforms(self) -> SubroutineUniformMap:
        '''
            SubroutineUniformMap: The subroutine uniforms of the program stage.
        '''

        return SubroutineUniformMap.new(self.mglo.subroutine_uniforms)



'''
    Mock linters.
'''

# pylint: skip-file

__all__ = ['mgl', 'Error']


class Error(Exception):
    '''
        ModernGL Error
    '''


class MockObject:
    '''
        MockObject
    '''

    wrapper = None
    mglo = None


class Implementation:
    '''
        Implementation
    '''

    BLEND = MockObject()
    DEPTH_TEST = MockObject()
    CULL_FACE = MockObject()
    MULTISAMPLE = MockObject()

    LINEAR = MockObject()
    NEAREST = MockObject()
    MIPMAP = MockObject()

    TRIANGLES = MockObject()
    TRIANGLE_STRIP = MockObject()
    TRIANGLE_FAN = MockObject()
    LINES = MockObject()
    LINE_STRIP = MockObject()
    LINE_LOOP = MockObject()
    POINTS = MockObject()
    LINE_STRIP_ADJACENCY = MockObject()
    LINES_ADJACENCY = MockObject()
    TRIANGLE_STRIP_ADJACENCY = MockObject()
    TRIANGLES_ADJACENCY = MockObject()

    def create_context(self) -> 'Context':
        '''
            create_context
        '''

    def create_standalone_context(self, *args) -> 'Context':
        '''
            create_standalone_context
        '''


mgl = Implementation()



'''
    ModernGL context
'''

# pylint: disable=too-many-public-methods

from typing import Tuple

try:
    from . import mgl
except ImportError:
    from .mock import mgl

from .common import InvalidObject
from .buffers import Buffer, detect_format
from .programs import ComputeShader, Shader, Program
from .vertex_arrays import VertexArray
from .textures import Texture, Texture3D
from .renderbuffers import Renderbuffer
from .framebuffers import Framebuffer


class Context:
    '''
        Create a :py:class:`Context` using:

        - :py:func:`ModernGL.create_context`
        - :py:func:`ModernGL.create_standalone_context`

        The :py:func:`create_context` must be used when rendering in a window.
        The :py:func:`create_standalone_context` must be used when rendering
        without a window.

        Members:

            - :py:meth:`Context.program`
            - :py:meth:`Context.vertex_shader`
            - :py:meth:`Context.fragment_shader`
            - :py:meth:`Context.geometry_shader`
            - :py:meth:`Context.tess_evaluation_shader`
            - :py:meth:`Context.tess_control_shader`
            - :py:meth:`Context.buffer`
            - :py:meth:`Context.simple_vertex_array`
            - :py:meth:`Context.vertex_array`
            - :py:meth:`Context.texture`
            - :py:meth:`Context.depth_texture`
            - :py:meth:`Context.renderbuffer`
            - :py:meth:`Context.depth_renderbuffer`
            - :py:meth:`Context.framebuffer`
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = Context.__new__(Context)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<Context>'

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def line_width(self) -> float:
        '''
            float: Set the default line width.
        '''

        return self.mglo.line_width

    @line_width.setter
    def line_width(self, value):
        self.mglo.line_width = value

    @property
    def point_size(self) -> float:
        '''
            float: Set the default point size.
        '''

        return self.mglo.point_size

    @point_size.setter
    def point_size(self, value):
        self.mglo.point_size = value

    @property
    def viewport(self) -> Tuple[int, int, int, int]:
        '''
            tuple: The viewport.

            Reading this property may force the GPU to sync.
            Use this property to set the viewport only.
        '''

        return self.mglo.viewport

    @viewport.setter
    def viewport(self, value):
        self.mglo.viewport = tuple(value)

    @property
    def max_samples(self) -> int:
        '''
            int: The max samples.
        '''

        return self.mglo.max_samples

    @property
    def max_integer_samples(self) -> int:
        '''
            int: The max integer samples.
        '''

        return self.mglo.max_integer_samples

    @property
    def max_texture_units(self) -> int:
        '''
            int: The max texture units.
        '''

        return self.mglo.max_texture_units

    @property
    def default_texture_unit(self) -> int:
        '''
            int: The default texture unit.
        '''

        return self.mglo.default_texture_unit

    @default_texture_unit.setter
    def default_texture_unit(self, value):
        self.mglo.default_texture_unit = value

    @property
    def default_framebuffer(self) -> Framebuffer:
        '''
            Framebuffer: The default framebuffer.
        '''

        return Framebuffer.new(self.mglo.default_framebuffer)

    @property
    def wireframe(self) -> bool:
        '''
            bool: Wireframe settings for debugging.
        '''

        return self.mglo.wireframe

    @wireframe.setter
    def wireframe(self, value):
        self.mglo.wireframe = value

    @property
    def front_face(self) -> str:
        '''
            str: The front_face.
        '''

        return self.mglo.front_face

    @front_face.setter
    def front_face(self, value):
        self.mglo.front_face = str(value)

    @property
    def error(self) -> str:
        '''
            str: The result of glGetError() but human readable.
            This values is provided for debug purposes only.
        '''

        return self.mglo.error

    @property
    def vendor(self) -> str:
        '''
            str: The vendor.
        '''

        return self.mglo.vendor

    @property
    def renderer(self) -> str:
        '''
            str: The renderer.
        '''

        return self.mglo.renderer

    @property
    def version(self) -> str:
        '''
            str: The OpenGL version.
        '''

        return self.mglo.version

    @property
    def version_code(self) -> int:
        '''
            int: The OpenGL version.
        '''

        return self.mglo.version_code

    @property
    def info(self) -> dict:
        '''
            dict: The result of multiple glGet.
        '''

        return self.mglo.info

    def clear(self, red=0.0, green=0.0, blue=0.0, alpha=0.0, *, viewport=None) -> None:
        '''
            Clear the framebuffer.

            Values must be in ``(0, 255)`` range.
            If the `viewport` is not ``None`` then scrissor test
            will be used to clear the given viewport.

            If the `viewport` is a 2-tuple it will clear the
            ``(0, 0, width, height)`` where ``(width, height)`` is the 2-tuple.

            If the `viewport` is a 4-tuple it will clear the given viewport.

            Args:
                red (float): color component.
                green (float): color component.
                blue (float): color component.
                alpha (float): alpha component.

            Keyword Args:
                viewport (tuple): The viewport.
        '''

        if viewport is not None:
            viewport = tuple(viewport)

        self.mglo.clear(red, green, blue, alpha, viewport)

    def enable(self, flag) -> None:
        '''
            Enable flags.

            Valid flags are:

                - :py:data:`ModernGL.BLEND`
                - :py:data:`ModernGL.DEPTH_TEST`
                - :py:data:`ModernGL.CULL_FACE`
                - :py:data:`ModernGL.MULTISAMPLE`

            Args:
                flag (EnableFlag): The flag to enable.
        '''

        self.mglo.enable(flag.mglo)

    def disable(self, flag) -> None:
        '''
            Disable flags.

            Valid flags are:

                - :py:data:`ModernGL.BLEND`
                - :py:data:`ModernGL.DEPTH_TEST`
                - :py:data:`ModernGL.CULL_FACE`
                - :py:data:`ModernGL.MULTISAMPLE`

            Args:
                flag (EnableFlag): The flag to disable.
        '''

        self.mglo.disable(flag.mglo)

    def finish(self) -> None:
        '''
            Wait for all drawing commands to finish.
        '''

        self.mglo.finish()

    def copy_buffer(self, dst, src, size=-1, *, read_offset=0, write_offset=0) -> None:
        '''
            Copy buffer content.

            Args:
                dst (Buffer): Destination buffer.
                src (Buffer): Source buffer.
                size (int): Size to copy.

            Keyword Args:
                read_offset (int): Read offset.
                write_offset (int): Write offset.
        '''

        self.mglo.copy_buffer(dst.mglo, src.mglo, size, read_offset, write_offset)

    def copy_framebuffer(self, dst, src) -> None:
        '''
            Copy framebuffer content.

            Use this method to:

                - blit framebuffers.
                - copy framebuffer content into a texture.
                - downsample framebuffers. (it will allow to read the framebuffer's content)
                - downsample a framebuffer directly to a texture.

            Args:
                dst (Framebuffer or Texture): Destination framebuffer or texture.
                src (Framebuffer): Source framebuffer.
        '''

        self.mglo.copy_framebuffer(dst.mglo, src.mglo)

    def buffer(self, data=None, *, reserve=0, dynamic=False) -> Buffer:
        '''
            Create a :py:class:`Buffer`.

            Args:
                data (bytes): Content of the new buffer.

            Keyword Args:
                reserve (int): The number of bytes to reserve.
                dynamic (bool): Treat buffer as dynamic.

            Returns:
                Buffer: buffer
        '''

        return Buffer.new(self.mglo.buffer(data, reserve, dynamic))

    def texture(self, size, components, data=None, *, samples=0, alignment=1, floats=False) -> Texture:
        '''
            Create a :py:class:`Texture`.

            Args:
                size (tuple): The width and height of the texture.
                components (int): The number of components 1, 2, 3 or 4.
                data (bytes): Content of the texture.

            Keyword Args:
                samples (int): The number of samples. Value `0` means no multisample format.
                alignment (int): The byte alignment 1, 2, 4 or 8.
                floats (bool): Use floating point precision.

            Returns:
                Texture: texture
        '''

        return Texture.new(self.mglo.texture(size, components, data, samples, alignment, floats))

    def texture3d(self, size, components, data=None, *, alignment=1, floats=False) -> Texture3D:
        '''
            Create a :py:class:`Texture3D`.

            Args:
                size (tuple): The width, height and depth of the texture.
                components (int): The number of components 1, 2, 3 or 4.
                data (bytes): Content of the texture.

            Keyword Args:
                alignment (int): The byte alignment 1, 2, 4 or 8.
                floats (bool): Use floating point precision.

            Returns:
                Texture3D: texture
        '''

        return Texture3D.new(self.mglo.texture3d(size, components, data, alignment, floats))

    def depth_texture(self, size, data=None, *, samples=0, alignment=4) -> Texture:
        '''
            Create a :py:class:`Texture`.

            Args:
                size (tuple): The width and height of the texture.
                data (bytes): Content of the texture.

            Keyword Args:
                samples (int): The number of samples. Value `0` means no multisample format.
                alignment (int): The byte alignment 1, 2, 4 or 8.

            Returns:
                Texture: depth texture
        '''

        return Texture.new(self.mglo.depth_texture(size, data, samples, alignment))

    def vertex_array(self, program, content, index_buffer=None) -> VertexArray:
        '''
            Create a :py:class:`VertexArray`.

            Args:
                program (Program): The program used by :py:meth:`~VertexArray.render` and
                                   :py:meth:`~VertexArray.transform`.
                content (list): A list of (buffer, format, attributes).
                index_buffer (Buffer): An index buffer.

            Returns:
                VertexArray: vertex array
        '''

        if index_buffer is not None:
            index_buffer = index_buffer.mglo

        content = tuple((a.mglo, b, tuple(c)) for a, b, c in content)
        return VertexArray.new(self.mglo.vertex_array(program.mglo, content, index_buffer))

    def simple_vertex_array(self, program, buffer, attributes) -> VertexArray:
        '''
            Create a :py:class:`VertexArray`.

            This is an alias for::

                format = detect_format(program, attributes)
                vertex_array(program, [(buffer, format, attributes)])

            Args:
                program (Program): The program used by :py:meth:`~VertexArray.render` and
                                   :py:meth:`~VertexArray.transform`.
                buffer (Buffer): The buffer.
                attributes (list): A list of attribute names.

            Returns:
                VertexArray: vertex array
        '''

        content = [(buffer, detect_format(program, attributes), attributes)]
        return self.vertex_array(program, content, None)

    def program(self, shaders, varyings=()) -> Program:
        '''
            Create a :py:class:`Program` object.

            Only linked programs will be returned.

            For more information please see: :py:class:`Program` and :py:class:`Shader`

            A single shader in the `shaders` parameter is also accepted.
            The varyings are only used when a transform program is created.

            Args:
                shaders (list): A list of :py:class:`Shader` objects.
                varyings (list): A list of varying names.

            Returns:
                Program: program

            Examples:

                A simple program designed for rendering::

                    >>> my_render_program = ctx.program([
                    ... 	ctx.vertex_shader(\'\'\'
                    ... 		#version 330
                    ...
                    ... 		in vec2 vert;
                    ...
                    ... 		void main() {
                    ... 			gl_Position = vec4(vert, 0.0, 1.0);
                    ... 		}
                    ... 	\'\'\'),
                    ... 	ctx.fragment_shader(\'\'\'
                    ... 		#version 330
                    ...
                    ... 		out vec4 color;
                    ...
                    ... 		void main() {
                    ... 			color = vec4(0.3, 0.5, 1.0, 1.0);
                    ... 		}
                    ... 	\'\'\'),
                    ... ])

                A simple program designed for transforming::

                    >>> my_vertex_shader = ctx.vertex_shader(\'\'\'
                    ...     #version 330
                    ...
                    ...     in vec4 vert;
                    ...     out float vert_length;
                    ...
                    ...     void main() {
                    ...         vert_length = length(vert);
                    ...     }
                    ... \'\'\')

                    >>> my_transform_program = ctx.program(my_vertex_shader, ['vert_length'])
        '''

        if isinstance(shaders, Shader):
            shaders = (shaders.mglo,)

        else:
            shaders = tuple(x.mglo for x in shaders)

        if isinstance(varyings, str):
            varyings = (varyings,)

        else:
            varyings = tuple(x for x in varyings)

        return Program.new(self.mglo.program(shaders, varyings))

    def vertex_shader(self, source) -> Shader:
        '''
            The Vertex Shader is a programmable Shader stage in the rendering pipeline
            that handles the processing of individual vertices.

            Vertex shaders are fed Vertex Attribute data, as specified from a vertex array
            object by a drawing command. A vertex shader receives a single vertex from the
            vertex stream and generates a single vertex to the output vertex stream.

            Args:
                source (str): The source code in GLSL.

            Returns:
                Shader: vertex shader

            Examples:

                Create a simple vertex shader::

                    >>> my_vertex_shader = ctx.vertex_shader(\'\'\'
                    ...     #version 330
                    ...
                    ...     in vec2 vert;
                    ...
                    ...     void main() {
                    ...         gl_Position = vec4(vert, 0.0, 1.0);
                    ...     }
                    ... \'\'\')
        '''

        return Shader.new(self.mglo.vertex_shader(source))

    def fragment_shader(self, source) -> Shader:
        '''
            A Fragment Shader is the Shader stage that will process a Fragment generated by
            the Rasterization into a set of colors and a single depth value.

            Args:
                source (str): The source code in GLSL.

            Returns:
                Shader: fragment shader

            Examples:

                Create a simple fragment shader::

                    >>> my_fragment_shader = ctx.fragment_shader(\'\'\'
                    ...     #version 330
                    ...
                    ...     out vec4 color;
                    ...
                    ...     void main() {
                    ...         color = vec4(0.3, 0.5, 1.0, 1.0);
                    ...     }
                    ... \'\'\')
        '''

        return Shader.new(self.mglo.fragment_shader(source))

    def geometry_shader(self, source) -> Shader:
        '''
            A Geometry Shader is a Shader program written in GLSL that governs the processing of Primitives.
            Geometry shaders reside between the Vertex Shaders (or the optional Tessellation stage) and the
            fixed-function Vertex Post-Processing stage.

            A geometry shader is optional and does not have to be used.

            Args:
                source (str): The source code in GLSL.

            Returns:
                Shader: geometry shader
        '''

        return Shader.new(self.mglo.geometry_shader(source))

    def tess_evaluation_shader(self, source) -> Shader:
        '''
            Tessellation is the Vertex Processing stage in the OpenGL rendering pipeline where
            patches of vertex data are subdivided into smaller Primitives.

            The Tessellation Evaluation Shader takes the tessellated patch and computes
            the vertex values for each generated vertex.

            Args:
                source (str): The source code in GLSL.

            Returns:
                Shader: tesselation evaluation shader
        '''

        return Shader.new(self.mglo.tess_evaluation_shader(source))

    def tess_control_shader(self, source) -> Shader:
        '''
            The Tessellation Control Shader (TCS) determines how much tessellation to do.
            It can also adjust the actual patch data, as well as feed additional patch data to later stages.
            The Tessellation Control Shader is optional.

            Args:
                source (str): The source code in GLSL.

            Returns:
                Shader: tesselation control shader
        '''

        return Shader.new(self.mglo.tess_control_shader(source))

    def framebuffer(self, color_attachments, depth_attachment=None) -> Framebuffer:
        '''
            A :py:class:`Framebuffer` is a collection of buffers that can be used as the destination for rendering.
            The buffers for Framebuffer objects reference images from either Textures or Renderbuffers.

            Args:
                color_attachments (list): A list of `Texture` or `Renderbuffer` objects.
                depth_attachment (Renderbuffer or Texture): A `Texture` or `Renderbuffer` object.

            Returns:
                Framebuffer: framebuffer
        '''

        if isinstance(color_attachments, (Texture, Renderbuffer)):
            color_attachments = (color_attachments.mglo,)

        else:
            color_attachments = tuple(x.mglo for x in color_attachments)

        if depth_attachment is not None:
            depth_attachment = depth_attachment.mglo

        return Framebuffer.new(self.mglo.framebuffer(color_attachments, depth_attachment))

    def renderbuffer(self, size, components=4, *, samples=0, floats=False) -> Renderbuffer:
        '''
            :py:class:`Renderbuffer` objects are OpenGL objects that contain images.
            They are created and used specifically with :py:class:`Framebuffer` objects.

            Args:
                size (tuple): The width and height of the renderbuffer.
                components (int): The number of components 1, 2, 3 or 4.

            Keyword Args:
                samples (int): The number of samples. Value `0` means no multisample format.
                floats (bool): Use floating point precision.

            Returns:
                Renderbuffer: renderbuffer
        '''

        return Renderbuffer.new(self.mglo.renderbuffer(size, components, samples, floats))

    def depth_renderbuffer(self, size, *, samples=0) -> Renderbuffer:
        '''
            :py:class:`Renderbuffer` objects are OpenGL objects that contain images.
            They are created and used specifically with :py:class:`Framebuffer` objects.

            Args:
                size (tuple): The width and height of the renderbuffer.

            Keyword Args:
                samples (int): The number of samples. Value `0` means no multisample format.

            Returns:
                Renderbuffer: depth renderbuffer
        '''

        return Renderbuffer.new(self.mglo.depth_renderbuffer(size, samples))

    def compute_shader(self, source) -> ComputeShader:
        '''
            A :py:class:`ComputeShader` is a Shader Stage that is used entirely for computing arbitrary information.
            While it can do rendering, it is generally used for tasks not directly related to drawing.

            Args:
                source (str): The source of the compute shader.

            Returns:
                ComputeShader: compute shader program
        '''

        return ComputeShader.new(self.mglo.compute_shader(source))

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject


def create_context(require=None) -> Context:
    '''
        Create a ModernGL context by loading OpenGL functions from an existing OpenGL context.
        An OpenGL context must exists. If rendering is done without a window please use the
        :py:func:`create_standalone_context` instead.

        Keyword Arguments:
            require (Version): OpenGL version.

        Returns:
            Context: context
    '''

    ctx = Context.new(mgl.create_context())

    if require is not None and ctx.version_code < require.code:
        raise Exception('The version required is not provided')

    return ctx


def create_standalone_context(size=(256, 256), require=None) -> Context:
    '''
        Create a standalone ModernGL context.
        This method will create a hidden window with the
        initial size given in the parameters. This will
        set the  initial viewport as well.

        It is reccommanded to use a separate framebuffer when rendering.

        The size is not really important for the default framebuffer.
        It is only useful to get a viewport with a visible size.
        Size ``(1, 1)`` is great to save memory, however makes harder to
        detect when the viewport was forgotten to set.

        Keyword Arguments:
            size (tuple): Initial framebuffer size.
            require (Version): OpenGL version.

        Returns:
            Context: context
    '''

    ctx = Context.new(mgl.create_standalone_context(*size))

    if require is not None and ctx.version_code < require.code:
        raise Exception('The version required is not provided')

    return ctx



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_texture_create_1(self):
        self.ctx.texture((16, 16), 3)

    def test_texture_create_2(self):
        pixels = b'\x10\x20\x30' * 16 * 16
        self.ctx.texture((16, 16), 3, pixels)

    def test_texture_create_string(self):
        pixels = 'abc' * 16 * 16
        with self.assertRaises(Exception):
            self.ctx.texture((16, 16), 3, pixels)

    def test_multisample_texture(self):
        if self.ctx.max_samples < 2:
            self.skipTest('multisampling is not supported')

        self.ctx.texture((16, 16), 3, samples=2)

    def test_depth_texture(self):
        self.ctx.depth_texture((16, 16))

    def test_multisample_depth_texture(self):
        if self.ctx.max_samples < 2:
            self.skipTest('multisampling is not supported')

        self.ctx.depth_texture((16, 16), samples=2)

    def test_texture_invalid_samples(self):
        if self.ctx.max_samples < 2:
            self.skipTest('multisampling is not supported')

        with self.assertRaisesRegex(ModernGL.Error, 'sample'):
            self.ctx.texture((16, 16), 3, samples=3)

    def test_texture_get_swizzle(self):
        tex = self.ctx.texture((16, 16), 4)
        self.assertEqual(tex.swizzle, 'RGBA')

    def test_texture_swizzle_1(self):
        tex = self.ctx.texture((16, 16), 4)
        tex.swizzle = 'argb'
        self.assertEqual(tex.swizzle, 'ARGB')

    def test_texture_swizzle_2(self):
        tex = self.ctx.texture((16, 16), 1)
        tex.swizzle = 'RRRR'
        self.assertEqual(tex.swizzle, 'RRRR')

    def test_texture_swizzle_3(self):
        tex = self.ctx.texture((16, 16), 2)
        tex.swizzle = '01RG'
        self.assertEqual(tex.swizzle, '01RG')

    def test_texture_read(self):
        pixels = b'\x10\x20\x30' * 8 * 8
        tex = self.ctx.texture((8, 8), 3, pixels)
        self.assertEqual(tex.read(), pixels)

    def test_texture_read_into(self):
        pixels = b'\x10\x20\x30' * 8 * 8
        tex = self.ctx.texture((8, 8), 3, pixels)
        buf = bytearray(8 * 8 * 3)
        tex.read_into(buf)
        self.assertEqual(bytes(buf), pixels)

    def test_texture_read_into_pbo(self):
        pixels = b'\x10\x20\x30' * 8 * 8
        tex = self.ctx.texture((8, 8), 3, pixels)
        buf = self.ctx.buffer(reserve=8 * 8 * 3)

        tex.read_into(buf)
        self.assertEqual(buf.read(), pixels)

    def test_texture_write_1(self):
        pixels1 = b'\x00\x00\x00' * 8 * 8
        pixels2 = b'\xff\xff\xff' * 8 * 8

        tex = self.ctx.texture((8, 8), 3, pixels1)
        self.assertEqual(tex.read(), pixels1)

        tex.write(pixels2)
        self.assertEqual(tex.read(), pixels2)

    def test_texture_write_2(self):
        pixels1 = b'\x00\x00\x00' * 8 * 8
        pixels2 = b'\xff\xff\xff' * 6 * 6
        pixels3 = b'\x10\x20\x30'

        tex = self.ctx.texture((8, 8), 3)

        tex.write(pixels1, viewport=(0, 0, 8, 8))
        tex.write(pixels2, viewport=(1, 1, 6, 6))
        tex.write(pixels3, viewport=(2, 4, 1, 1))

        def pixel(x, y):
            if x == 2 and y == 4:
                return b'\x10\x20\x30'

            if x == 0 or y == 0 or x == 7 or y == 7:
                return b'\x00\x00\x00'

            return b'\xff\xff\xff'

        expectation = b''.join(pixel(x, y) for y in range(8) for x in range(8))
        self.assertEqual(tex.read(), expectation)

    def test_texture_alignment_1(self):
        tex = self.ctx.texture((3, 3), 3)
        self.assertEqual(len(tex.read(alignment=1)), 27)

    def test_texture_alignment_2(self):
        tex = self.ctx.texture((3, 3), 3)
        self.assertEqual(len(tex.read(alignment=2)), 30)

    def test_texture_alignment_3(self):
        tex = self.ctx.texture((3, 3), 3)
        self.assertEqual(len(tex.read(alignment=4)), 36)

    def test_texture_alignment_4(self):
        tex = self.ctx.texture((3, 3), 3)
        self.assertEqual(len(tex.read(alignment=8)), 48)

    def test_texture_alignment_5(self):
        pixels = b'\x80\x90\xA0\x80\x90\xA0\x80\x90\xA0\x80\x90\xA0'
        tex = self.ctx.texture((2, 2), 3, pixels, alignment=1)
        self.assertEqual(tex.read(alignment=1), pixels)

    def test_texture_alignment_6(self):
        pixels = b'\x80\x90\xA0\x80\x90\xA0\x00\x00\x80\x90\xA0\x80\x90\xA0\x00\x00'
        tex = self.ctx.texture((2, 2), 3, pixels, alignment=4)
        result = tex.read(alignment=4)
        self.assertEqual(result[0:6], b'\x80\x90\xA0\x80\x90\xA0')
        self.assertEqual(result[8:14], b'\x80\x90\xA0\x80\x90\xA0')


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_renderbuffer(self):
        self.ctx.renderbuffer((64, 64))

    def test_multisample_renderbuffer(self):
        if self.ctx.max_samples < 2:
            self.skipTest('multisampling is not supported')

        self.ctx.renderbuffer((64, 64), samples=2)

    def test_depth_renderbuffer(self):
        self.ctx.depth_renderbuffer((64, 64))

    def test_multisample_depth_renderbuffer(self):
        if self.ctx.max_samples < 2:
            self.skipTest('multisampling is not supported')

        self.ctx.depth_renderbuffer((64, 64), samples=2)

    def test_renderbuffer_invalid_samples(self):
        with self.assertRaisesRegex(ModernGL.Error, 'sample'):
            self.ctx.renderbuffer((64, 64), samples=3)


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_vertex_shader(self):
        src = '''
            #version 330
            in vec4 vertex;
            void main() {
                gl_Position = vertex;
            }
        '''

        shader = self.ctx.vertex_shader(src)
        self.assertEqual(shader.source, src)
        self.assertEqual(shader.typename, 'vertex_shader')

    def test_fragment_shader(self):
        src = '''
            #version 330
            out vec4 color;
            void main() {
                color = vec4(0.0, 0.0, 0.0, 1.0);
            }
        '''

        shader = self.ctx.fragment_shader(src)
        self.assertEqual(shader.source, src)
        self.assertEqual(shader.typename, 'fragment_shader')

    def test_geometry_shader(self):
        src = '''
            #version 330
            layout(points) in;
            layout(points, max_vertices = 1) out;
            out vec4 color;
            void main() {
                gl_Position = gl_in[0].gl_Position;
                EmitVertex();
                EndPrimitive();
            }
        '''

        shader = self.ctx.geometry_shader(src)
        self.assertEqual(shader.source, src)
        self.assertEqual(shader.typename, 'geometry_shader')

    def test_tess_control_shader(self):
        if self.ctx.version_code < 400:
            self.skipTest('OpenGL 4.0 is not supported')

        src = '''
            #version 400
            layout(vertices = 3) out;
            in vec3 vPosition[];
            out vec3 tcPosition[];
            uniform float TessLevelInner;
            uniform float TessLevelOuter;
            void main() {
                tcPosition[gl_InvocationID] = vPosition[gl_InvocationID];
                if (gl_InvocationID == 0) {
                    gl_TessLevelInner[0] = TessLevelInner;
                    gl_TessLevelOuter[0] = TessLevelOuter;
                    gl_TessLevelOuter[1] = TessLevelOuter;
                    gl_TessLevelOuter[2] = TessLevelOuter;
                }
            }
        '''

        shader = self.ctx.tess_control_shader(src)
        self.assertEqual(shader.source, src)
        self.assertEqual(shader.typename, 'tess_control_shader')

    def test_tess_evaluation_shader(self):
        if self.ctx.version_code < 400:
            self.skipTest('OpenGL 4.0 is not supported')

        src = '''
            #version 400
            layout(triangles, equal_spacing, cw) in;
            in vec3 tcPosition[];
            out vec3 tePosition;
            out vec3 tePatchDistance;
            uniform mat4 Projection;
            uniform mat4 Modelview;
            void main() {
                vec3 p0 = gl_TessCoord.x * tcPosition[0];
                vec3 p1 = gl_TessCoord.y * tcPosition[1];
                vec3 p2 = gl_TessCoord.z * tcPosition[2];
                tePatchDistance = gl_TessCoord;
                tePosition = normalize(p0 + p1 + p2);
                gl_Position = Projection * Modelview * vec4(tePosition, 1);
            }
        '''

        shader = self.ctx.tess_evaluation_shader(src)
        self.assertEqual(shader.source, src)
        self.assertEqual(shader.typename, 'tess_evaluation_shader')


if __name__ == '__main__':
    unittest.main()



'''
    ModernGL main
'''

import argparse
import json
import sys

import ModernGL


def main():
    '''
        main
    '''

    parser = argparse.ArgumentParser(prog='ModernGL')
    parser.add_argument('-v', '--version', action='version', version='ModernGL %s' % ModernGL.__version__)
    parser.add_argument('-i', '--info', action='store_true', default=False)
    args = parser.parse_args()

    ctx = ModernGL.create_standalone_context()

    if args.info:
        print(json.dumps(ctx.info, sort_keys=True, indent=4))

    else:
        print('ModernGL:', ModernGL.__version__)
        print('Vendor:', ctx.vendor)
        print('Renderer:', ctx.renderer)
        print('Version:', ctx.version)
        print('Python:', sys.version)
        print('Platform:', sys.platform)


if __name__ == '__main__':
    main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_1(self):
        if self.ctx.version_code < 430:
            self.skipTest('OpenGL 4.3 is not supported')

        compute_shader = self.ctx.compute_shader('''
            #version 430

            layout (local_size_x = 1, local_size_y = 1) in;

            layout (std430, binding = 1) buffer Input {
                float v1[4];
            };

            layout (std430, binding = 2) buffer Output {
                float v2[4];
            };

            void main() {
                v2[0] = v1[3];
                v2[1] = v1[2];
                v2[2] = v1[1];
                v2[3] = v1[0];
            }
        ''')

        buf1 = self.ctx.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))
        buf2 = self.ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))

        buf1.bind_to_storage_buffer(1)
        buf2.bind_to_storage_buffer(2)

        compute_shader.run()

        a, b, c, d = struct.unpack('4f', buf2.read())

        self.assertAlmostEqual(a, 4.0)
        self.assertAlmostEqual(b, 3.0)
        self.assertAlmostEqual(c, 2.0)
        self.assertAlmostEqual(d, 1.0)


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()
        cls.vbo = cls.ctx.buffer(reserve=1024)
        cls.res = cls.ctx.buffer(reserve=1024)
        cls.vbo.clear()

        cls.shader = '''
            #version %(version)s
            uniform %(type)s Uniform;
            in %(type)s v_in;
            out %(type)s v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        '''

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def test_float_uniform(self):
        vertex_shader = self.ctx.vertex_shader('''
            #version 330
            uniform float Uniform;
            in float v_in;
            out float v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''')

        prog = self.ctx.program(vertex_shader, ['v_out'])
        vbo = self.ctx.buffer(struct.pack('f', 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, ['v_in'])

        prog.uniforms['Uniform'].value = 7.0
        vao.transform(self.res)

        val = struct.unpack('f', self.res.read(4))[0]
        self.assertAlmostEqual(val, 7.0)

    def test_int_uniform(self):
        vertex_shader = self.ctx.vertex_shader('''
            #version 330
            uniform int Uniform;
            in int v_in;
            out int v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''')

        prog = self.ctx.program(vertex_shader, ['v_out'])
        vbo = self.ctx.buffer(struct.pack('i', 1))
        vao = self.ctx.simple_vertex_array(prog, vbo, ['v_in'])

        prog.uniforms['Uniform'].value = -2
        vao.transform(self.res)

        val = struct.unpack('i', self.res.read(4))[0]
        self.assertAlmostEqual(val, -2)

    def test_vec_uniform(self):
        vertex_shader = self.ctx.vertex_shader('''
            #version 330
            uniform vec3 Uniform;
            in vec3 v_in;
            out vec3 v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''')

        prog = self.ctx.program(vertex_shader, ['v_out'])
        vbo = self.ctx.buffer(struct.pack('3f', 1.0, 1.0, 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, ['v_in'])

        prog.uniforms['Uniform'].value = (0.5, 1.0, 1.5)
        vao.transform(self.res)

        x, y, z = struct.unpack('3f', self.res.read(12))
        self.assertAlmostEqual(x, 0.5)
        self.assertAlmostEqual(y, 1.0)
        self.assertAlmostEqual(z, 1.5)

    def test_mat_uniform(self):
        vertex_shader = self.ctx.vertex_shader('''
            #version 330
            uniform mat2x3 Uniform;
            in float v_in;
            out mat2x3 v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''')

        prog = self.ctx.program(vertex_shader, ['v_out'])
        vbo = self.ctx.buffer(struct.pack('f', 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, ['v_in'])

        prog.uniforms['Uniform'].value = (0.0, 1.0, 2.0, 3.0, 4.0, 5.0)
        vao.transform(self.res)

        m = struct.unpack('6f', self.res.read(24))
        self.assertAlmostEqual(m[0], 0.0)
        self.assertAlmostEqual(m[1], 1.0)
        self.assertAlmostEqual(m[2], 2.0)
        self.assertAlmostEqual(m[3], 3.0)
        self.assertAlmostEqual(m[4], 4.0)
        self.assertAlmostEqual(m[5], 5.0)


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        cls.vertex_shader = cls.ctx.vertex_shader('''
            #version 330
            uniform mat4 Mvp;
            in vec4 vertex;
            uniform Block {
                mat4 Transform;
            };
            void main() {
                gl_Position = Mvp * Transform * vertex;
            }
        ''')

        cls.geometry_shader = cls.ctx.geometry_shader('''
            #version 330
            layout(points) in;
            uniform vec4 Position;
            layout(line_strip, max_vertices = 2) out;
            out vec4 color;
            void main() {
                gl_Position = gl_in[0].gl_Position + Position;
                EmitVertex();
                gl_Position = gl_in[0].gl_Position - Position;
                EmitVertex();
                EndPrimitive();
            }
        ''')

        cls.fragment_shader = cls.ctx.fragment_shader('''
            #version 330
            uniform vec4 Color;
            uniform Lights {
                uniform vec4 Light[1];
            };
            out vec4 color;
            void main() {
                color = Color * Light[0];
            }
        ''')

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_program_uniforms_1(self):
        program = self.ctx.program(self.vertex_shader)

        self.assertIn('Mvp', program.uniforms)
        self.assertNotIn('Block', program.uniforms)
        self.assertNotIn('Transform', program.uniforms)

    def test_program_uniforms_2(self):
        program = self.ctx.program([self.vertex_shader, self.geometry_shader, self.fragment_shader])

        self.assertIn('Mvp', program.uniforms)
        self.assertIn('Position', program.uniforms)
        self.assertIn('Color', program.uniforms)

    def test_program_uniform_blocks(self):
        program = self.ctx.program([self.vertex_shader, self.fragment_shader])

        self.assertIn('Block', program.uniform_blocks)
        self.assertIn('Lights', program.uniform_blocks)
        self.assertNotIn('Transform', program.uniform_blocks)
        self.assertNotIn('Light', program.uniform_blocks)
        self.assertNotIn('Block', program.uniforms)
        self.assertNotIn('Lights', program.uniforms)

    def test_program_geometry_primitives_1(self):
        program = self.ctx.program(self.vertex_shader)

        self.assertIsNone(program.geometry_input)
        self.assertIsNone(program.geometry_output)
        self.assertEqual(program.geometry_vertices, 0)

    def test_program_geometry_primitives_2(self):
        program = self.ctx.program([self.vertex_shader, self.geometry_shader])

        self.assertIsInstance(program.geometry_input, ModernGL.Primitive)
        self.assertIsInstance(program.geometry_output, ModernGL.Primitive)
        self.assertEqual(program.geometry_vertices, 2)

    def test_program_duplicate_shader_1(self):
        with self.assertRaisesRegex(ModernGL.Error, 'duplicate'):
            self.ctx.program([self.vertex_shader, self.vertex_shader])

    def test_program_duplicate_shader_2(self):
        with self.assertRaisesRegex(ModernGL.Error, 'duplicate'):
            self.ctx.program([self.vertex_shader, self.fragment_shader, self.fragment_shader])


if __name__ == '__main__':
    unittest.main()



'''
    ModernGL common
'''

# pylint: disable=too-few-public-methods

try:
    from . import mgl
except ImportError:
    from .mock import mgl


class InvalidObject:
    '''
        A ModernGL object turns into an InvalidObject
        once the ``release()`` method is successfully called.
    '''

    __slots__ = ['mglo']


class EnableFlag:
    '''
        EnableFlag
    '''

    __slots__ = ['mglo', 'name']

    @staticmethod
    def new(obj, name) -> 'EnableFlag':
        '''
            For internal use only.
        '''

        res = EnableFlag.__new__(EnableFlag)
        res.name = name
        res.mglo = obj
        return res

    def __init__(self):
        self.name = None
        self.mglo = None
        raise NotImplementedError('EnableFlag')

    def __repr__(self):
        return 'ModernGL.%s' % self.name

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


BLEND = EnableFlag.new(mgl.BLEND, 'BLEND')
'''
    GL_BLEND
'''

DEPTH_TEST = EnableFlag.new(mgl.DEPTH_TEST, 'DEPTH_TEST')
'''
    GL_DEPTH_TEST
'''

CULL_FACE = EnableFlag.new(mgl.CULL_FACE, 'CULL_FACE')
'''
    GL_CULL_FACE
'''

MULTISAMPLE = EnableFlag.new(mgl.MULTISAMPLE, 'MULTISAMPLE')
'''
    GL_MULTISAMPLE
'''


class Primitive:
    '''
        Primitive
    '''

    __slots__ = ['mglo', 'name']

    @staticmethod
    def new(obj, name) -> 'Primitive':
        '''
            For internal use only.
        '''

        res = Primitive.__new__(Primitive)
        obj.wrapper = res
        res.name = name
        res.mglo = obj
        return res

    def __init__(self):
        self.name = None
        self.mglo = None
        raise NotImplementedError('Primitive')

    def __repr__(self):
        return 'ModernGL.%s' % self.name

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


TRIANGLES = Primitive.new(mgl.TRIANGLES, 'TRIANGLES')
'''
    GL_TRIANGLES
'''

TRIANGLE_STRIP = Primitive.new(mgl.TRIANGLE_STRIP, 'TRIANGLE_STRIP')
'''
    GL_TRIANGLE_STRIP
'''

TRIANGLE_FAN = Primitive.new(mgl.TRIANGLE_FAN, 'TRIANGLE_FAN')
'''
    GL_TRIANGLE_FAN
'''

LINES = Primitive.new(mgl.LINES, 'LINES')
'''
    GL_LINES
'''

LINE_STRIP = Primitive.new(mgl.LINE_STRIP, 'LINE_STRIP')
'''
    GL_LINE_STRIP
'''

LINE_LOOP = Primitive.new(mgl.LINE_LOOP, 'LINE_LOOP')
'''
    GL_LINE_LOOP
'''

POINTS = Primitive.new(mgl.POINTS, 'POINTS')
'''
    GL_POINTS
'''

LINE_STRIP_ADJACENCY = Primitive.new(mgl.LINE_STRIP_ADJACENCY, 'LINE_STRIP_ADJACENCY')
'''
    GL_LINE_STRIP_ADJACENCY
'''

LINES_ADJACENCY = Primitive.new(mgl.LINES_ADJACENCY, 'LINES_ADJACENCY')
'''
    GL_LINES_ADJACENCY
'''

TRIANGLE_STRIP_ADJACENCY = Primitive.new(mgl.TRIANGLE_STRIP_ADJACENCY, 'TRIANGLE_STRIP_ADJACENCY')
'''
    GL_TRIANGLE_STRIP_ADJACENCY
'''

TRIANGLES_ADJACENCY = Primitive.new(mgl.TRIANGLES_ADJACENCY, 'TRIANGLES_ADJACENCY')
'''
    GL_TRIANGLES_ADJACENCY
'''


class Version:
    '''
        Version
    '''

    __slots__ = ['_major', '_minor']

    def __init__(self, major, minor):
        self._major = major
        self._minor = minor

    def __repr__(self):
        return 'ModernGL.CORE_%d' % self.code

    def __eq__(self, other):
        return self.major == other.major and self.minor == other.minor

    def __ne__(self, other):
        return self.major != other.major or self.minor != other.minor

    @property
    def major(self) -> int:
        '''
            int: major
        '''

        return self._major

    @property
    def minor(self) -> int:
        '''
            int: minor
        '''

        return self._minor

    @property
    def code(self) -> int:
        '''
            int: The version code.

            The version code is similar to the GLSL version specification.
            This value is always equal to `major * 100 + minor * 10`.
            The purpose of this proeprty is to quickly compare the versions.
        '''

        return self._major * 100 + self._minor * 10


CORE_330 = Version(3, 3)
'''
    OpenGL 3.3
'''

CORE_400 = Version(4, 0)
'''
    OpenGL 4.0
'''

CORE_410 = Version(4, 1)
'''
    OpenGL 4.1
'''

CORE_420 = Version(4, 2)
'''
    OpenGL 4.2
'''

CORE_430 = Version(4, 3)
'''
    OpenGL 4.3
'''

CORE_440 = Version(4, 4)
'''
    OpenGL 4.4
'''

CORE_450 = Version(4, 5)
'''
    OpenGL 4.5
'''


class TextureFilter:
    '''
        TextureFilter
    '''

    __slots__ = ['mglo', 'name']

    @staticmethod
    def new(obj, name) -> 'TextureFilter':
        '''
            For internal use only.
        '''

        res = TextureFilter.__new__(TextureFilter)
        obj.wrapper = res
        res.name = name
        res.mglo = obj
        return res

    def __init__(self):
        self.name = None
        self.mglo = None
        raise NotImplementedError('TextureFilter')

    def __repr__(self):
        return 'ModernGL.%s' % self.name

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo


LINEAR = TextureFilter.new(mgl.LINEAR, 'LINEAR')
'''
    (GL_LINEAR, GL_LINEAR)
'''

NEAREST = TextureFilter.new(mgl.NEAREST, 'NEAREST')
'''
    (GL_NEAREST, GL_NEAREST)
'''

MIPMAP = TextureFilter.new(mgl.MIPMAP, 'MIPMAP')
'''
    (GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR)
'''



'''
    ModernGL vertex arrays
'''

# pylint: disable=too-few-public-methods

from typing import Tuple

from .common import InvalidObject, TRIANGLES, POINTS
from .programs import Program
from .buffers import Buffer


class VertexArrayAttribute:
    '''
        VertexArrayAttribute
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = VertexArrayAttribute.__new__(VertexArrayAttribute)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError('VertexArrayAttributeMap')

    def __getitem__(self, key) -> 'VertexArrayAttribute':
        return VertexArrayAttribute.new(self.mglo[key])

    @property
    def location(self):
        '''
            location
        '''

        return self.mglo.location

    @property
    def divisor(self):
        '''
            divisor
        '''

        return self.mglo.divisor

    @property
    def stride(self):
        '''
            stride
        '''

        return self.mglo.stride

    @property
    def enabled(self):
        '''
            enabled
        '''

        return self.mglo.enabled

    @enabled.setter
    def enabled(self, value):
        self.mglo.enabled = value

    def bind(self, buffer, offset=0, stride=0, divisor=0) -> None:
        '''
            bind
        '''

        self.mglo.bind(buffer.mglo, offset, stride, divisor)

    def enable(self) -> None:
        '''
            enable
        '''

        self.mglo.enable()

    def disable(self) -> None:
        '''
            disable
        '''

        self.mglo.disable()


class VertexArrayAttributeMap:
    '''
        VertexArrayAttributeMap
    '''

    __slots__ = ['mglo']

    def __init__(self):
        self.mglo = {}
        raise NotImplementedError('VertexArrayAttributeMap')

    def __getitem__(self, key) -> VertexArrayAttribute:
        return VertexArrayAttribute.new(self.mglo[key])

    def __contains__(self, key):
        return key in self.mglo

    def __iter__(self):
        yield from self.mglo.items()

    def __len__(self):
        return len(self.mglo)

    def __repr__(self):
        return repr(self.mglo)

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = VertexArrayAttributeMap.__new__(VertexArrayAttributeMap)
        res.mglo = obj
        return res


class VertexArray:
    '''
        A VertexArray object is an OpenGL object that stores all of the state
        needed to supply vertex data. It stores the format of the vertex data
        as well as the Buffer objects providing the vertex data arrays.

        In ModernGL, the VertexArray object also stores a reference
        for a :py:class:`Program` object, and some Subroutine information.

        A VertexArray object cannot be instantiated directly, it requires a context.
        Use :py:meth:`Context.vertex_array` or :py:meth:`Context.simple_vertex_array`
        to create one.
    '''

    __slots__ = ['mglo']

    @staticmethod
    def new(obj):
        '''
            For internal use only.
        '''

        res = VertexArray.__new__(VertexArray)
        res.mglo = obj
        return res

    def __init__(self):
        self.mglo = None
        raise NotImplementedError()

    def __repr__(self):
        return '<VertexArray: %d>' % self.glo

    def __eq__(self, other):
        return self.mglo is other.mglo

    def __ne__(self, other):
        return self.mglo is not other.mglo

    @property
    def program(self) -> Program:
        '''
            Program: The program assinged to the VertexArray.
            The program used when rendering or transforming primitives.
        '''

        return Program.new(self.mglo.program)

    @property
    def attributes(self) -> VertexArrayAttributeMap:
        '''
            VertexArrayAttributeMap: Individual vertex attributes.
            Use the :py:meth:`VertexArrayAttribute.bind` method
            to assign vertex attributes to buffers.
        '''

        return VertexArrayAttributeMap.new(self.mglo.attributes)

    @property
    def index_buffer(self) -> Buffer:
        '''
            Buffer: The index buffer if the index_buffer is set, otherwise ``None``.
        '''

        return self.mglo.index_buffer

    @property
    def vertices(self) -> int:
        '''
            int: The number of vertices detected.
            This is the minimum of the number of vertices possible per Buffer.
            The size of the index_buffer determines the number of vertices.
            Per instance vertex attributes does not affect this number.
        '''

        return self.mglo.vertices

    @property
    def subroutines(self) -> Tuple[int]:
        '''
            tuple: The subroutines assinged to the VertexArray.
            The subroutines used when rendering or transforming primitives.
        '''

        return self.mglo.subroutines

    @subroutines.setter
    def subroutines(self, value):
        self.mglo.subroutines = tuple(value)

    @property
    def glo(self) -> int:
        '''
            int: The internal OpenGL object.
            This values is provided for debug purposes only.
        '''

        return self.mglo.glo

    def render(self, mode=TRIANGLES, vertices=-1, *, first=0, instances=1) -> None:
        '''
            The render primitive (mode) must be the same as
            the input primitive of the GeometryShader.

            Args:
                mode (Primitive): By default `TRIANGLES` will be used.
                vertices (int): The number of vertices to transform.

            Keyword Args:
                first (int): The index of the first vertex to start with.
                instances (int): The number of instances.
        '''

        self.mglo.render(mode.mglo, vertices, first, instances)

    def transform(self, buffer, mode=POINTS, vertices=-1, *, first=0, instances=1) -> None:
        '''
            Transform vertices.
            Stores the output in a single buffer.
            The transform primitive (mode) must be the same as
            the input primitive of the GeometryShader.

            Args:
                buffer (Buffer): The buffer to store the output.
                mode (Primitive): By default `TRIANGLES` will be used.
                vertices (int): The number of vertices to transform.

            Keyword Args:
                first (int): The index of the first vertex to start with.
                instances (int): The number of instances.
        '''

        self.mglo.transform(buffer.mglo, mode.mglo, vertices, first, instances)

    def release(self) -> None:
        '''
            Release the ModernGL object.
        '''

        self.mglo.release()
        self.__class__ = InvalidObject



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_buffer_readonly_access(self):
        buf = self.ctx.buffer(b'Hello World!')
        with buf.access(offset=1, size=4, readonly=True) as acc:
            self.assertEqual(acc.offset, 1)
            self.assertEqual(acc.size, 4)
            self.assertTrue(acc.readonly)
            self.assertEqual(acc.read(), b'ello')

    def test_buffer_readwrite_access(self):
        buf = self.ctx.buffer(b'Hello World!')
        with buf.access(offset=6, size=5, readonly=False) as acc:
            self.assertEqual(acc.offset, 6)
            self.assertEqual(acc.size, 5)
            self.assertFalse(acc.readonly)
            self.assertEqual(acc.read(), b'World')
            acc.write(b'ORLD', offset=1)
            self.assertEqual(acc.read(offset=2), b'RLD')


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def test_sampler_2D_getsetter(self):
        self.prog = self.ctx.program(self.ctx.vertex_shader('''
            #version 330
            uniform sampler2D Texture2D;
            void main() {
                gl_Position = texture(Texture2D, vec2(0.0, 0.0));
            }
        '''))

        sampler_uniform = self.prog.uniforms['Texture2D']

        sampler_uniform.value = 3
        self.assertEqual(sampler_uniform.value, 3)

        sampler_uniform.value = 1
        self.assertEqual(sampler_uniform.value, 1)

    def test_sampler_3D_getsetter(self):
        self.prog = self.ctx.program(self.ctx.vertex_shader('''
            #version 330
            uniform sampler3D Texture3D;
            void main() {
                gl_Position = texture(Texture3D, vec3(0.0, 0.0, 0.0));
            }
        '''))

        sampler_uniform = self.prog.uniforms['Texture3D']

        sampler_uniform.value = 3
        self.assertEqual(sampler_uniform.value, 3)

        sampler_uniform.value = 1
        self.assertEqual(sampler_uniform.value, 1)


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestBuffer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        cls.vert = cls.ctx.vertex_shader('''
            #version 330

            in vec2 in_v1;
            in vec2 in_v2;

            out vec2 out_v;

            void main() {
                out_v = in_v1 + in_v2;
            }
        ''')

        cls.prog = cls.ctx.program(cls.vert, ['out_v'])

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_1(self):
        buf_v1 = self.ctx.buffer(struct.pack('8f', 1, 2, 3, 4, 5, 6, 7, 8))
        buf_v2 = self.ctx.buffer(struct.pack('2f', 10, 100))
        res = self.ctx.buffer(reserve=buf_v1.size)

        vao = self.ctx.vertex_array(self.prog, [
            (buf_v1, '2f', ['in_v1']),
            (buf_v2, '2f/i', ['in_v2']),
        ])

        vao.transform(res, ModernGL.POINTS)
        tup = struct.unpack('8f', res.read())
        self.assertAlmostEqual(tup[0], 11.0)
        self.assertAlmostEqual(tup[1], 102.0)
        self.assertAlmostEqual(tup[2], 13.0)
        self.assertAlmostEqual(tup[3], 104.0)
        self.assertAlmostEqual(tup[4], 15.0)
        self.assertAlmostEqual(tup[5], 106.0)
        self.assertAlmostEqual(tup[6], 17.0)
        self.assertAlmostEqual(tup[7], 108.0)


if __name__ == '__main__':
    unittest.main()



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_1(self):
        if self.ctx.version_code < 400:
            self.skipTest('OpenGL 4.0 is not supported')

        vbo1 = self.ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))
        vbo2 = self.ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))

        vert = self.ctx.vertex_shader('''
            #version 400

            in vec4 vert;
            out vec4 color;

            subroutine vec4 color_t();

            subroutine(color_t)
            vec4 ColorRed() {
                return vec4(1, 0, 0, 1);
            }

            subroutine(color_t)
            vec4 ColorBlue() {
                return vec4(0, 0.5, 1, 1);
            }

            subroutine(color_t)
            vec4 ColorYellow() {
                return vec4(1, 1, 0, 1);
            }

            subroutine uniform color_t Color;

            void main() {
                color = vert + Color();
            }
        ''')

        prog = self.ctx.program(vert, ['color'])
        vao = self.ctx.simple_vertex_array(prog, vbo1, ['vert'])

        vao.subroutines = (prog.vertex_shader.subroutines['ColorRed'],)
        vao.transform(vbo2)

        x, y, z, w = struct.unpack('4f', vbo2.read())
        self.assertAlmostEqual(x, 1.0)
        self.assertAlmostEqual(y, 0.0)
        self.assertAlmostEqual(z, 0.0)
        self.assertAlmostEqual(w, 1.0)

        vao.subroutines = [prog.vertex_shader.subroutines['ColorBlue']]
        vao.transform(vbo2)

        x, y, z, w = struct.unpack('4f', vbo2.read())
        self.assertAlmostEqual(x, 0.0)
        self.assertAlmostEqual(y, 0.5)
        self.assertAlmostEqual(z, 1.0)
        self.assertAlmostEqual(w, 1.0)

        vao.subroutines = [prog.vertex_shader.subroutines['ColorYellow'].index]
        vao.transform(vbo2)

        x, y, z, w = struct.unpack('4f', vbo2.read())
        self.assertAlmostEqual(x, 1.0)
        self.assertAlmostEqual(y, 1.0)
        self.assertAlmostEqual(z, 0.0)
        self.assertAlmostEqual(w, 1.0)


if __name__ == '__main__':
    unittest.main()



import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_transform(self):

        vertex_shader = self.ctx.vertex_shader('''
            #version 330

            in vec4 vert;
            out float vert_length;

            void main() {
                vert_length = length(vert);
            }
        ''')

        program = self.ctx.program(vertex_shader, ['vert_length'])

        self.assertIsNot(program.vertex_shader, None)
        self.assertIs(program.fragment_shader, None)
        self.assertIs(program.geometry_shader, None)
        self.assertIs(program.tess_evaluation_shader, None)
        self.assertIs(program.tess_control_shader, None)

        self.assertIs(program.geometry_input, None)
        self.assertIs(program.geometry_output, None)

        self.assertEqual(len(program.uniforms), 0)
        self.assertEqual(len(program.attributes), 1)

        self.assertEqual(len(program.varyings), 1)
        self.assertEqual(len(program.uniform_blocks), 0)

        self.assertTrue('vert' in program.attributes)
        self.assertTrue('vert' not in program.varyings)

        self.assertTrue('vert_length' not in program.attributes)
        self.assertTrue('vert_length' in program.varyings)


if __name__ == '__main__':
    unittest.main()



import glob
import platform
import sys

from setuptools import Extension, setup

target = platform.system().lower()

if target == 'linux':
    from distutils import sysconfig
    cvars = sysconfig.get_config_vars()

    if 'OPT' in cvars:
        sysconfig._config_vars['OPT'] = cvars['OPT'].replace('-Wstrict-prototypes', '')

    if 'CFLAGS' in cvars:
        sysconfig._config_vars['CFLAGS'] = cvars['CFLAGS'].replace('-Wstrict-prototypes', '')

install_requires = []

if sys.version_info < (3, 5):
    install_requires.append('typing')

libraries = {
    'windows': ['gdi32', 'opengl32', 'user32'],
    'linux': ['GL', 'dl', 'X11'],
    'darwin': [],
}

extra_compile_args = {
    'windows': [],
    'linux': [],
    'darwin': ['-Wno-deprecated-declarations'],
}

extra_linker_args = {
    'windows': [],
    'linux': [],
    'darwin': ['-framework', 'OpenGL', '-Wno-deprecated'],
}

ModernGL = Extension(
    name='ModernGL.mgl',
    include_dirs=['src'],
    define_macros=[
        # ('MGL_DEBUG', None),
        # ('MGL_VERBOSE', None),
    ],
    libraries=libraries[target],
    extra_compile_args=extra_compile_args[target],
    extra_link_args=extra_linker_args[target],
    sources=glob.glob('src/*.cpp'),
)

short_description = 'ModernGL: PyOpenGL alternative'

long_description = '''
`ModernGL on github <https://github.com/cprogrammer1994/ModernGL>`_

`Documentation <https://moderngl.readthedocs.io/>`_

`Examples <https://moderngl.github.io/Examples.html>`_

OpenGL is a great environment for developing portable, platform independent,
interactive 2D and 3D graphics applications. The API implementation in Python
is cumbersome, resulting in applications with high latency. To solve this
problem we have developed ModernGL, a wrapper over OpenGL that simplifies the
creation of simple graphics applications like scientific simulations, small
games or user interfaces. Usually, acquiring in-depth knowledge of OpenGL
requires a steep learning curve. In contrast, ModernGL is easy to learn and
use, moreover it is capable of rendering with the same performance and quality,
with less code written.
'''

keywords = [
    'ModernGL',
    'modern OpenGL',
    'OpenGL',
    'PyOpenGL',
    'visualization',
    'ray-tracing',
    'compute shader',
    'shader',
    'documentation',
    'graphics',
    'GLSL',
    'GPU',
    'GPGPU',
    'nvidia',
    'amd',
    'GL',
    'GLU',
    'GLEXT',
    'WGL',
    'WGLEXT',
    'ARB',
    'GLX',
    '2D',
    '3D',
    'CAD',
    'design',
    'video',
]

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: MacOS X',
    'Environment :: Win32 (MS Windows)',
    'Environment :: X11 Applications',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: OS Independent',
    'Operating System :: Unix',
    'Topic :: Artistic Software',
    'Topic :: Desktop Environment',
    'Topic :: Documentation :: Sphinx',
    'Topic :: Games/Entertainment',
    'Topic :: Multimedia',
    'Topic :: Multimedia :: Graphics',
    'Topic :: Multimedia :: Graphics :: 3D Modeling',
    'Topic :: Multimedia :: Graphics :: 3D Rendering',
    'Topic :: Multimedia :: Video :: Display',
    'Topic :: Scientific/Engineering :: Visualization',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3 :: Only',
]

args = {
    'name': 'ModernGL',
    'version': '4.1.11',
    'description': short_description,
    'long_description': long_description.strip(),
    'url': 'https://github.com/cprogrammer1994/ModernGL',
    'download_url': 'https://github.com/cprogrammer1994/ModernGL/releases',
    'author': 'Szabolcs Dombi',
    'author_email': 'cprogrammer1994@gmail.com',
    'license': 'MIT',
    'classifiers': classifiers,
    'keywords': keywords,
    'packages': ['ModernGL', 'ModernGL.ext', 'ModernGL.tk'],
    'install_requires': install_requires,
    'ext_modules': [ModernGL],
    'platforms': ['any'],
}

if target == 'windows':
    args['zip_safe'] = False

setup(**args)



import struct
import unittest

import ModernGL


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = ModernGL.create_standalone_context()

        cls.vert = cls.ctx.vertex_shader('''
            #version 330

            in int a_in;
            in int b_in;
            in int c_in;
            in int d_in;

            out int a_out;
            out int b_out;
            out int c_out;
            out int d_out;

            void main() {
                a_out = a_in * 2;
                b_out = b_in * 2;
                c_out = c_in * 2;
                d_out = d_in * 2;
            }
        ''')

        cls.prog = cls.ctx.program(cls.vert, ['a_out', 'b_out', 'c_out', 'd_out'])

    @classmethod
    def tearDownClass(cls):
        cls.ctx.release()

    def tearDown(self):
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')

    def test_padding_1(self):
        buf = self.ctx.buffer(struct.pack('=ixi12xii', 1, 2, 3, 4))
        res = self.ctx.buffer(reserve=16)

        vao = self.ctx.vertex_array(self.prog, [
            (buf, 'ixi12xii', ['a_in', 'b_in', 'c_in', 'd_in']),
        ])

        vao.transform(res, ModernGL.POINTS)
        a, b, c, d = struct.unpack('=iiii', res.read())

        self.assertEqual(a, 2)
        self.assertEqual(b, 4)
        self.assertEqual(c, 6)
        self.assertEqual(d, 8)

    def test_padding_2(self):
        buf = self.ctx.buffer(struct.pack('=i8xi8xi8xi8x', 1, 2, 3, 4))
        res = self.ctx.buffer(reserve=64)

        vao = self.ctx.vertex_array(self.prog, [
            (buf, 'i8x', ['a_in']),
        ])

        vao.transform(res, ModernGL.POINTS)

        self.assertEqual(vao.vertices, 4)

        a1, a2, a3, a4 = struct.unpack('=i12xi12xi12xi12x', res.read())

        self.assertEqual(a1, 2)
        self.assertEqual(a2, 4)
        self.assertEqual(a3, 6)
        self.assertEqual(a4, 8)

    def test_padding_3(self):
        buf = self.ctx.buffer(struct.pack('=128xiiii', 1, 2, 3, 4))
        res = self.ctx.buffer(reserve=16)

        vao = self.ctx.vertex_array(self.prog, [
            (buf, '128xiiii', ['a_in', 'b_in', 'c_in', 'd_in']),
        ])

        vao.transform(res, ModernGL.POINTS)
        a, b, c, d = struct.unpack('=iiii', res.read())

        self.assertEqual(a, 2)
        self.assertEqual(b, 4)
        self.assertEqual(c, 6)
        self.assertEqual(d, 8)


if __name__ == '__main__':
    unittest.main()
